/* --------
   Includes
   -------- */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>


/* ---------------------------------
   Global variables are defined here
   --------------------------------- */

struct global_data *GLOBAL;
struct local_data *LOCAL;
#ifdef YAPOR
struct local_data *REMOTE[MAX_WORKERS];
struct worker WORKER;
#endif /* YAPOR */


/* ----------------
   Global functions
   ---------------- */

void abort_optyap(const char *msg, ...) {
  va_list args;

  va_start(args, msg);
  fprintf(stderr, "[ ");
#ifdef YAPOR
  fprintf (stderr, "Worker %d ", worker_id);
#endif /* YAPOR */
  fprintf (stderr, "Aborting OPTYap -> ");
  vfprintf(stderr, msg, args);
  fprintf(stderr, " ]\n");

  exit_optyap();
  exit (1);
}


int condit_binds, shared_binds, uncond_binds;

void exit_optyap(void) {
  char MapFile[20];
#ifdef YAPOR
  int proc;
  if (GLOBAL) {
    for (proc = 0; proc < number_workers; proc++) {
      if (proc != worker_id && worker_pid(proc) != 0) {
        if (kill(worker_pid(proc), SIGKILL) != 0)
          optyap_message_type0("Can't kill process %d", worker_pid(proc));
        else 
          optyap_message_type0("Killing process %d", worker_pid(proc));
      }
    }
  }
#endif /* YAPOR */
  strcpy(MapFile,"/tmp/mapfile");
#ifdef YAPOR
#ifdef DEBUG
printf("worker %d did %d shared, %d conditional, %d uncond\n", worker_id, shared_binds, condit_binds, uncond_binds);
#endif
  if (worker_id == 0)
    itos(getpid(), &MapFile[12]);
  else itos(worker_pid(0), &MapFile[12]);
#else /* TABLING */
  itos(getpid(), &MapFile[12]);
#endif /* YAPOR */
  if (remove(MapFile) == 0)
    optyap_message_type0("Removing mapfile \"%s\"", MapFile);
  else optyap_message_type0("Can't remove mapfile \"%s\"", MapFile);
  return;
}


void itos(int i, char *s) {
  int n,r,j;
  n = 10;
  while (n <= i) n *= 10;
  j = 0;
  while (n > 1) {
    n = n / 10;   
    r = i / n;
    i = i - r * n;
    s[j++] = r + '0';
  }
  s[j] = 0;
  return;
}


void optyap_print_message(const char *s,...) {
  va_list args;
  va_start(args, s);
  fprintf(stderr, "[ ");
#ifdef YAPOR
  fprintf(stderr, "W%d: ", worker_id);
#endif /* YAPOR */
  vfprintf(stderr, s, args);
  fprintf(stderr, " ]\n");
  return;
}
#endif /* YAPOR || TABLING */
