/* -------------------------------------
   Includes, defines and local variables
   ------------------------------------- */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include <stdio.h>
#include "Yatom.h"
#ifdef YAPOR
#include "or.macros.h"
#endif /* YAPOR */
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */

#ifdef YAPOR
#define ANSWER_YES -1
#define ANSWER_NO -2
static int length_answer;
static ans_fr_ptr actual_answer;
#endif /* YAPOR */


/* ---------------------------
   Local functions declaration
   --------------------------- */

#ifdef YAPOR
static int p_default_sequential(void);
static int p_execution_mode(void);
static int p_sequential(void);
static int p_parallel_new_answer(void);
static int p_parallel_answer_yes(void);
static int p_parallel_answer_no(void);
static int parallel_new_answer_putchar(int ch);
static void show_answers(void);
static void answer_to_stdout(char *answer);
#endif /* YAPOR */
#ifdef TABLING
static int p_tabling_present(void);
static int p_table(void);
static int p_show_trie(void);
#endif /* TABLING */
static int p_show_frames(void);


/* ----------------
   Global functions
   ---------------- */

void init_optyap_preds(void) {
#ifdef YAPOR
  InitCPred("$start_yapor", 0, yapor_on, SafePredFlag);
  InitCPred("$default_sequential", 1, p_default_sequential, SafePredFlag);
  InitCPred("$sequential", 1, p_sequential, SafePredFlag);
  InitCPred("$parallel_new_answer", 1, p_parallel_new_answer, SafePredFlag);
  InitCPred("$parallel_answer_yes", 0, p_parallel_answer_yes, SafePredFlag);
  InitCPred("$parallel_answer_no", 0, p_parallel_answer_no, SafePredFlag);
  InitCPred("performance", 1, p_performance, SafePredFlag);
  InitCPred("execution_mode", 1, p_execution_mode, SafePredFlag);
#endif /* YAPOR */
#ifdef TABLING
  InitCPred("$table", 1, p_table, SafePredFlag);
  InitCPred("$show_trie", 1, p_show_trie, SafePredFlag);
  InitCPred("$tabling_present", 0, p_tabling_present, SafePredFlag);
  InitCPred("$abolish_incomplete_tables", 0, p_abolish_incomplete_tables, SafePredFlag);
  InitCPred("$check_incomplete_tables", 0, p_check_incomplete_tables, SafePredFlag);
#endif /* TABLING */
  InitCPred("show_frames", 0, p_show_frames, SafePredFlag);
}


#ifdef YAPOR
int yapor_on(void) {
  if (GLOBAL_execution_mode == SEQUENTIAL_MODE) 
    return (FALSE);
  START_PARALLEL_GOAL(worker_id);
  while (!BITMAP_empty(GLOBAL_bm_finished_workers)) continue;
  GLOBAL_execution_time = current_time();
  SIMICS_SERVICE(START_COUNTER, 0);
  SIMICS_SERVICE(START_COUNTER, 1);
  SIMICS_ATOMIC_SERVICE(ONE_MORE_TASK);
#ifdef TABLING
  reset_suspension_area();
#endif /* TABLING */
  make_root_choice_point();   
  BITMAP_delete(GLOBAL_bm_idle_workers, 0);
  BITMAP_clear(GLOBAL_bm_invisible_workers);
  BITMAP_clear(GLOBAL_bm_requestable_workers);
  GLOBAL_answers = 0;
  GLOBAL_parallel_mode = EXECUTING_MODE;
  return (TRUE);
}


void yapor_off(void) {
  while (!BITMAP_same(GLOBAL_bm_present_workers, GLOBAL_bm_finished_workers)) continue;
  SIMICS_SERVICE(STOP_COUNTER, 5);
  SIMICS_SERVICE(STOP_COUNTER, 0);
#ifdef SIMICS
  exit_optyap();
  exit (1);
#endif /* SIMICS */
  GLOBAL_execution_time = current_time() - GLOBAL_execution_time;
  switch(GLOBAL_answers) {
    case ANSWER_YES: fprintf(stdout, "[ yes (in %f seconds) ]\n\n", GLOBAL_execution_time);
                     break;
    case ANSWER_NO:  fprintf(stdout, "[ no (in %f seconds) ]\n\n", GLOBAL_execution_time);
                     break;
    default:         show_answers();
                     break;
  }
  GLOBAL_parallel_mode = SUSPENDED_MODE;
  return;
}
#endif /* YAPOR */


/* ---------------
   Local functions
   --------------- */

#ifdef YAPOR
static int p_default_sequential(void) {
  Term t;
  char *s;
  t = Deref(ARG1);
  if (IsVarTerm(t)) {
    Term ta;
    if (SEQUENTIAL_IS_DEFAULT) ta = MkAtomTerm(LookupAtom("on"));
    else ta = MkAtomTerm(LookupAtom("off"));
    Bind((CELL *)t, ta);
    return(TRUE);
  } 
  if (!IsAtomTerm(t)) 
    return(FALSE);
  s = RepAtom(AtomOfTerm(t))->StrOfAE;
  if (strcmp(s, "on") == 0) {
    SEQUENTIAL_IS_DEFAULT = TRUE;
    return(TRUE);
  } 
  if (strcmp(s,"off") == 0) {
    SEQUENTIAL_IS_DEFAULT = FALSE;
    return(TRUE);
  }
  return(FALSE);
}


static int p_execution_mode(void) {
  Term t;
  char *s;
  t = Deref(ARG1);
  if (IsVarTerm(t)) {
    Term ta;
    if (GLOBAL_execution_mode == SEQUENTIAL_MODE) 
      ta = MkAtomTerm(LookupAtom("sequential"));
    else 
      ta = MkAtomTerm(LookupAtom("parallel"));
    Bind((CELL *)t, ta);
    return(TRUE);
  } 
  if (!IsAtomTerm(t)) return(FALSE);
  s = RepAtom(AtomOfTerm(t))->StrOfAE;
  if (strcmp(s, "parallel") == 0) {
    GLOBAL_execution_mode = PARALLEL_MODE;
    return(TRUE);
  } 
  if (strcmp(s,"sequential") == 0) {
    GLOBAL_execution_mode = SEQUENTIAL_MODE;
    return(TRUE);
  }
  return(FALSE);
}


static int p_sequential(void) {
  Term t;
  Atom at;
  int arity;
  PredEntry *pe;

  t = Deref(ARG1);
  if (IsAtomTerm(t)) {
    at = AtomOfTerm(t);
    arity = 0;
  } else if (IsApplTerm(t)) {
    Functor func = FunctorOfTerm(t);
    at = NameOfFunctor(func);
    arity = ArityOfFunctor(func);
  } else {
    at = NULL; /* = 0 just to avoid gcc warning */
    arity = 0; /* = 0 just to avoid gcc warning */
    abort_optyap("unknown term in function p_sequential");
  }
  pe = RepPredProp(PredProp(at, arity));
  pe->PredFlags |= SequentialPredFlag;
  return (TRUE);
}


static int p_parallel_new_answer(void) {
  choiceptr leftmost_cp;
  or_fr_ptr leftmost_or_fr;

  ALLOC_ANSWER_FRAME(actual_answer);
  length_answer = 0;
  plwrite(ARG1, parallel_new_answer_putchar, 4);
  AnsFr_answer(actual_answer)[length_answer] = 0;
  AnsFr_next(actual_answer) = NULL;
  leftmost_cp = CUT_leftmost_until();
  leftmost_or_fr = leftmost_cp->cp_or_fr;
  LOCK_OR_FRAME(leftmost_or_fr);
  if (LOCAL_cut_cp != NULL) {
    FREE_ANSWER_FRAME(actual_answer);
  } else {      
    CUT_put_new_solution(leftmost_or_fr, actual_answer);
  }
  UNLOCK_OR_FRAME(leftmost_or_fr);
  return (TRUE);
}


static int p_parallel_answer_yes(void) {
  LOCK(GLOBAL_lock_free_answer_frame);
  GLOBAL_answers = ANSWER_YES;
  UNLOCK(GLOBAL_lock_free_answer_frame);
  return (TRUE);
}


static int p_parallel_answer_no(void) {
  LOCK(GLOBAL_lock_free_answer_frame);
  if (GLOBAL_answers != ANSWER_YES) 
    GLOBAL_answers = ANSWER_NO;
  UNLOCK(GLOBAL_lock_free_answer_frame);
  return (TRUE);
}


static int parallel_new_answer_putchar(int ch) {
  AnsFr_answer(actual_answer)[length_answer++] = ch;
  return ch;
}


static void show_answers(void) {
  int i;
  if (OrFr_solutions(LOCAL_top_or_fr) != NULL) {
    ans_fr_ptr aux_answer1, aux_answer2;
    aux_answer1 = SolFr_first(OrFr_solutions(LOCAL_top_or_fr));
    while (aux_answer1 != NULL) {
      answer_to_stdout(AnsFr_answer(aux_answer1));
      aux_answer2 = aux_answer1;
      aux_answer1 = AnsFr_next(aux_answer1);
      FREE_ANSWER_FRAME(aux_answer2);
      GLOBAL_answers++;
    }
    FREE_SOLUTION_FRAME(OrFr_solutions(LOCAL_top_or_fr));
    OrFr_solutions(LOCAL_top_or_fr) = NULL;
  }
  switch (GLOBAL_answers) {
    case 0:  fprintf(stdout, "[ no (in %f seconds) ]\n\n", GLOBAL_execution_time);
             break;
    case 1:  fprintf(stdout, "[ 1 answer founded (in %f seconds) ]\n\n", GLOBAL_execution_time);
             break;
    default: fprintf(stdout, "[ %d answers founded (in %f seconds) ]\n\n", GLOBAL_answers, GLOBAL_execution_time);
             break;
  }

  for (i = GLOBAL_number_goals; i > 0; i--) {
    if (GLOBAL_best_times(i) > GLOBAL_execution_time) {
      if (i + 1 < MAX_BEST_TIMES)
        GLOBAL_best_times(i + 1) = GLOBAL_best_times(i);
      else {
        GLOBAL_best_times(0) -= GLOBAL_best_times(i);
      }
    }
    else break;
  }
  if (i + 1 < MAX_BEST_TIMES) {
    GLOBAL_best_times(0) += GLOBAL_execution_time;
    GLOBAL_best_times(i + 1) = GLOBAL_execution_time;
    if (GLOBAL_number_goals + 1 < MAX_BEST_TIMES)
      GLOBAL_number_goals++;
  }

  return;
}


static void answer_to_stdout(char *answer) {
  int length_answer = 0, length_output = 0, caracter, list, par_rectos;
  char output[MAX_LENGTH_ANSWER];
  while (1) {
    length_answer += 2;
    while (answer[length_answer] != ']') {
      length_answer++;
      caracter = 0;
      while (answer[length_answer] != ',' && answer[length_answer] != ']')
	caracter = caracter * 10 + answer[length_answer++] - '0';
      output[length_output++] = caracter;
    }
    length_answer++;
    output[length_output++] = ' ';
    output[length_output++] = '=';
    output[length_output++] = ' '; 
    if (answer[length_answer++] == ',') {
      list = 1;
      output[length_output++] = '[';
    } else list = 0;
    par_rectos = 1;
    while (1) {
      if (answer[length_answer] == '[') par_rectos++;
      else if (answer[length_answer] == ']' && --par_rectos == 0) break;
      output[length_output++] = answer[length_answer++];
    }
    if (list) output[length_output++] = ']';
    if (answer[++length_answer] != ']') {
      output[length_output++] = ' ';
      output[length_output++] = ';';
      output[length_output++] = ' ';
    }
    else break;
  }
  output[length_output] = 0;
  fprintf(stdout, "  %s\n", output);
  return;
}
#endif /* YAPOR */


#ifdef TABLING
static int p_tabling_present(void) {
  return (TRUE);
}


static int p_table(void) {
  Term t;
  Atom at;
  int arity;
  PredEntry *pe;
  tab_ent_ptr te;
  sg_node_ptr sg_node;

  t = Deref(ARG1);
  if (IsAtomTerm(t)) {
    at = AtomOfTerm(t);
    arity = 0;
  } else if (IsApplTerm(t)) {
    Functor func = FunctorOfTerm(t);
    at = NameOfFunctor(func);
    arity = ArityOfFunctor(func);
  } else {
    at = NULL; /* = 0 just to avoid gcc warning */
    arity = 0; /* = 0 just to avoid gcc warning */
    abort_optyap("unknown term in function p_table"); 
  }
  pe = RepPredProp(PredProp(at, arity));
  pe->PredFlags |= TabledPredFlag;
  new_subgoal_trie_node(sg_node, 0, NULL, NULL, NULL, NULL);
  new_table_entry(te, sg_node);
  pe->TableOfPred = te;
  return (TRUE);
}


static int p_show_trie(void) {
  Term t;
  Atom at;
  int arity;
  PredEntry *pe;

  t = Deref(ARG1);
  if (IsAtomTerm(t)) {
    at = AtomOfTerm(t);
    arity = 0;
  } else if (IsApplTerm(t)) {
    Functor func = FunctorOfTerm(t);
    at = NameOfFunctor(func);
    arity = ArityOfFunctor(func);
  } else {
    at = NULL; /* = 0 just to avoid gcc warning */
    arity = 0; /* = 0 just to avoid gcc warning */
    abort_optyap("unknown term in function p_show_trie");
  }
  pe = RepPredProp(PredProp(at, arity));
  show_trie(TrNode_child(TabEnt_subgoal_trie(pe->TableOfPred)), arity, 1);
  return (TRUE);
}
#endif /* TABLING */


static int p_show_frames(void) {
  int cont;
#ifdef YAPOR
  or_fr_ptr aux1 = GLOBAL_free_or_frame;
  ans_fr_ptr aux2 = GLOBAL_free_answer_frame;
  sol_fr_ptr aux3 = GLOBAL_free_solution_frame;
#endif /* YAPOR */
#ifdef TABLING
  tab_ent_ptr aux4 = GLOBAL_free_table_entry;
  sg_node_ptr aux5 = GLOBAL_free_subgoal_trie_node;
  sg_fr_ptr aux6 = GLOBAL_free_subgoal_frame;
  ans_node_ptr aux7 = GLOBAL_free_answer_trie_node;
  dep_fr_ptr aux8 = GLOBAL_free_dependency_frame;
#ifdef YAPOR
  susp_fr_ptr aux9 = GLOBAL_free_suspension_frame;
#endif /* YAPOR */
#endif /* TABLING */

  fprintf(stdout, "[\n");
#ifdef YAPOR
  cont = 0;
  while (aux1 != NULL) {
    cont++;
    aux1 = OrFr_next(aux1);
  }
  fprintf(stdout, "  Free or frames: %d\n", cont);
  cont = 0;
  while (aux2 != NULL) {
    cont++;
    aux2 = AnsFr_next(aux2);
  }
  fprintf(stdout, "  Free answer frames: %d\n", cont);
  cont = 0;
  while (aux3 != NULL) {
    cont++;
    aux3 = SolFr_next(aux3);
  }
  fprintf(stdout, "  Free solution frames: %d\n", cont);
#endif /* YAPOR */
#ifdef TABLING
  cont = 0;
  while (aux4 != NULL) {
    cont++;
    aux4 = TabEnt_next(aux4);
  }
  fprintf(stdout, "  Free table entries: %d\n", cont);
  cont = 0;
  while (aux5 != NULL) {
    cont++;
    aux5 = TrNode_next(aux5);
  }
  fprintf(stdout, "  Free subgoal trie nodes: %d\n", cont);
  cont = 0;
  while (aux6 != NULL) {
    cont++;
    aux6 = SgFr_next(aux6);
  }
  fprintf(stdout, "  Free subgoal frames: %d\n", cont);
  cont = 0;
  while (aux7 != NULL) {
    cont++;
    aux7 = TrNode_next(aux7);
  }
  fprintf(stdout, "  Free answer trie nodes: %d\n", cont);
  cont = 0;
  while (aux8 != NULL) {
    cont++;
    aux8 = DepFr_next(aux8);
  }
  fprintf(stdout, "  Free dependency frames: %d\n", cont);
#ifdef YAPOR
  cont = 0;
  while (aux9 != NULL) {
    cont++;
    aux9 = SuspFr_next(aux9);
  }
  fprintf(stdout, "  Free suspension frames: %d\n", cont);
#endif /* YAPOR */
#endif /* TABLING */
  fprintf(stdout, "]\n\n");
  return (TRUE);
}
#endif /* YAPOR || TABLING */
