/* ---------------------------
   Prototype for opt.*.c files
   --------------------------- */

/* opt.memory.c */
int global_data_size(void);
int local_data_size(void);
void map_memory(int HeapArea, int GlobalLocalArea, int TrailAuxArea, int fd, int n_workers);
void alloc_all_optyap_structs(void);
#ifdef YAPOR
void remap_memory(int fd);
void alloc_or_frames(void);
void alloc_answer_frames(void);
void alloc_solution_frames(void);
#endif /* YAPOR */
#ifdef TABLING
void alloc_table_entries(void);
void alloc_subgoal_trie_nodes(void);
void alloc_subgoal_frames(void);
void alloc_answer_trie_nodes(void);
void alloc_dependency_frames(void);
#ifdef YAPOR
void alloc_suspension_frames(void);
void *alloc_suspension_block(int size);
void reset_suspension_area(void);
#endif /* YAPOR */
#endif /* TABLING */


/* opt.misc.c */
void abort_optyap(const char *msg, ...);
void exit_optyap(void);
void itos(int i, char *s);
void optyap_print_message(const char *s, ...);


/* opt.init.c */
void init_global(int n_workers, int sch_loop, int delay_load);
void init_local(void);
void make_root_frames(void);
void init_signals(void);
#ifdef YAPOR
void init_workers(int fd);
#endif /* YAPOR */


/* opt.preds.c */
void init_optyap_preds(void);
#ifdef YAPOR
int yapor_on(void);
void yapor_off(void);
#endif /* YAPOR */


/* opt.performance.c */
#ifdef YAPOR
realtime current_time(void);
int p_performance(void);
#endif /* YAPOR */


/* ---------------------------
   Prototype for tab.*.c files
   --------------------------- */

/* tab.tries.c */
#ifdef TABLING
sg_node_ptr subgoal_search(OPREG arity, sg_node_ptr current_sg_node, CELL **Yaddr);
ans_node_ptr answer_search(ans_node_ptr current_ans_node, CELL *subs_ptr);
void show_trie(sg_node_ptr sg_node, int missing_nodes, int level);
int p_abolish_incomplete_tables(void);
int p_check_incomplete_tables(void);
#endif /* TABLING */


/* tab.suspend.c */
#ifdef TABLING
void private_completion(void);
#ifdef YAPOR
void public_completion(void);
void suspend_completion_branch(void);
void resume_bottom_collected_suspension_branch(void);
void collect_suspension_frames_with_unconsumed_answers(void);
int suspension_frame_with_unconsumed_answers(susp_fr_ptr susp_fr);
#endif /* YAPOR */
#endif /* TABLING */


/* --------------------------
   Prototype for or.*.c files
   -------------------------- */

/* or.engine.c */
#ifdef YAPOR
void make_root_choice_point(void);
void free_root_choice_point(void);
int q_share_work(int p);
void p_share_work(void);
#endif /* YAPOR */


/* or.scheduler.c */
#ifdef YAPOR
int get_work(void);
#endif /* YAPOR */


/* or.cut.c */
#ifdef YAPOR
int no_workers_left_pending_cut(void);
void cut_shared(choiceptr cut_cp, sol_fr_ptr aux_solution);
#endif /* YAPOR */
