/* ------------------
   struct table_entry
   ------------------ */

typedef struct table_entry {
  struct subgoal_trie_node *subgoal_trie;
  struct table_entry *next;
} *tab_ent_ptr;

#define TabEnt_subgoal_trie(X)  ((X)->subgoal_trie)
#define TabEnt_next(X)          ((X)->next)



/* ----------------------------------------------
   structs subgoal_trie_node and answer_trie_node
   ---------------------------------------------- */

typedef struct subgoal_trie_node {
  lockvar lock;
  Term entry;
  struct subgoal_trie_node *child;
  struct subgoal_trie_node *parent;
  struct subgoal_trie_node *previous;
  struct subgoal_trie_node *next;
} *sg_node_ptr;

typedef struct answer_trie_node {
  OPCODE trie_instruction;         /* u.opc = opcode(op_instruction) */
  COUNT alternatives_left_to_try;  /* u.ld.ltt */
  COUNT have_cuts;                 /* u.ld.cut */
  COUNT is_sequential;             /* u.ld.seq */ 
  int op_instruction;
  lockvar lock;
  Term entry;
  struct answer_trie_node *child;
  struct answer_trie_node *parent;
  struct answer_trie_node *next;
} *ans_node_ptr;

#define TrNode_instr(X)     ((X)->trie_instruction)
#define TrNode_ltt(X)       ((X)->alternatives_left_to_try)
#define TrNode_cut(X)       ((X)->have_cuts)
#define TrNode_seq(X)       ((X)->is_sequential)
#define TrNode_op_instr(X)  ((X)->op_instruction)
#define TrNode_lock(X)      ((X)->lock)
#define TrNode_entry(X)     ((X)->entry)
#define TrNode_child(X)     ((X)->child)
#define TrNode_parent(X)    ((X)->parent)
#define TrNode_previous(X)  ((X)->previous)
#define TrNode_next(X)      ((X)->next)



/* --------------------
   struct subgoal_frame
   -------------------- */

typedef struct subgoal_frame {
  lockvar lock;
  struct generator_choice_point *generator_choice_point;
  struct subgoal_trie_node *subgoal_trie;
  struct answer_trie_node *answer_trie;
  struct answer_trie_node *first_answer;
  struct answer_trie_node *last_answer;
  int completed_flag;
#ifdef YAPOR
  int generator_worker;
  struct or_frame *generator_or_frame;
#endif /* YAPOR */
  struct subgoal_frame *next;
} *sg_fr_ptr;

#define SgFr_lock(X)          ((X)->lock)
#define SgFr_gen_cp(X)        ((X)->generator_choice_point)
#define SgFr_subgoal_trie(X)  ((X)->subgoal_trie)
#define SgFr_answer_trie(X)   ((X)->answer_trie)
#define SgFr_first_answer(X)  ((X)->first_answer)
#define SgFr_last_answer(X)   ((X)->last_answer)
#define SgFr_completed(X)     ((X)->completed_flag)
#define SgFr_gen_worker(X)    ((X)->generator_worker)
#define SgFr_gen_or_fr(X)     ((X)->generator_or_frame)
#define SgFr_next(X)          ((X)->next)



/* -----------------------
   struct dependency_frame
   ----------------------- */

typedef struct dependency_frame {
  lockvar lock;
  struct consumer_choice_point *consumer_choice_point;
  struct choicept *leader_choice_point;
  struct subgoal_frame *subgoal_frame;
  struct answer_trie_node *last_consumed_answer;
  CELL *h_freeze_register;
  struct choicept *b_freeze_register;
  struct trail_frame *tr_freeze_register;
#ifdef YAPOR
  struct choicept *top_chain_choice_point;
#endif /* YAPOR */
  struct dependency_frame *next;
} *dep_fr_ptr;

#define DepFr_lock(X)          ((X)->lock)
#define DepFr_cons_cp(X)       ((X)->consumer_choice_point)
#define DepFr_leader_cp(X)     ((X)->leader_choice_point)
#define DepFr_sg_fr(X)         ((X)->subgoal_frame)
#define DepFr_last_ans(X)      ((X)->last_consumed_answer)
#define DepFr_h_fz(X)          ((X)->h_freeze_register)
#define DepFr_b_fz(X)          ((X)->b_freeze_register)
#define DepFr_tr_fz(X)         ((X)->tr_freeze_register)
#define DepFr_top_chain_cp(X)  ((X)->top_chain_choice_point)
#define DepFr_next(X)          ((X)->next)



/* -----------------------
   struct suspension_frame
   ----------------------- */

#ifdef YAPOR
typedef struct suspension_frame {
  int suspending_worker;
  struct choicept *top_choice_point;
  struct choicept *top_consumer_choice_point;
  struct dependency_frame *top_dependency_frame;
  struct subgoal_frame *top_subgoal_frame;
  struct suspended_block {
    void *resume_register;
    void *block_start;
    int block_size;
  } global_block, local_block, trail_block;
  struct suspension_frame *next;
} *susp_fr_ptr;
#endif /* YAPOR */

#define SuspFr_susp_worker(X)      ((X)->suspending_worker)
#define SuspFr_top_cp(X)           ((X)->top_choice_point)
#define SuspFr_top_consumer_cp(X)  ((X)->top_consumer_choice_point)
#define SuspFr_top_dep_fr(X)       ((X)->top_dependency_frame)
#define SuspFr_top_sg_fr(X)        ((X)->top_subgoal_frame)
#define SuspFr_global_reg(X)       ((X)->global_block.resume_register)
#define SuspFr_global_start(X)     ((X)->global_block.block_start)
#define SuspFr_global_size(X)      ((X)->global_block.block_size)
#define SuspFr_local_reg(X)        ((X)->local_block.resume_register)
#define SuspFr_local_start(X)      ((X)->local_block.block_start)
#define SuspFr_local_size(X)       ((X)->local_block.block_size)
#define SuspFr_trail_reg(X)        ((X)->trail_block.resume_register)
#define SuspFr_trail_start(X)      ((X)->trail_block.block_start)
#define SuspFr_trail_size(X)       ((X)->trail_block.block_size)
#define SuspFr_next(X)             ((X)->next)



/* -------------------
   struct generator_cp
   ------------------- */

typedef struct generator_choice_point {
  /* common choicepoints fields */
  tr_fr_ptr gcp_tr;
  CELL *gcp_h;
  struct choicept *gcp_b;
#ifdef DEPTH_LIMIT
  CELL gcp_depth;
#endif /* DEPTH_LIMIT */
  yamop *gcp_cp;
#ifdef YAPOR
  int gcp_lub;
  struct or_frame *gcp_or_fr;
#endif /* YAPOR */
  yamop *gcp_ap;
  CELL *gcp_env;

  /* specific generator choicepoint field */
  struct subgoal_frame *gcp_sg_fr;
} *gen_cp_ptr;



/* ------------------
   struct consumer_cp
   ------------------ */

typedef struct consumer_choice_point {
  /* common choicepoints fields */
  tr_fr_ptr ccp_tr;
  CELL *ccp_h;
  struct choicept *ccp_b;
#ifdef DEPTH_LIMIT
  CELL ccp_depth;
#endif /* DEPTH_LIMIT */
  yamop *ccp_cp;
#ifdef YAPOR
  int ccp_lub;
  struct or_frame *ccp_or_fr;
#endif /* YAPOR */
  yamop *ccp_ap;
  CELL *ccp_env;

  /* specific consumer choicepoint field */
  struct dependency_frame *ccp_dep_fr;
} *cons_cp_ptr;
