/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef TABLING
#include "tab.macros.h"
#ifdef YAPOR
#include "or.macros.h"
#endif /* YAPOR */



/* ---------------------------
   Local functions declaration
   --------------------------- */

#ifdef YAPOR
void release_suspension_branch(susp_fr_ptr susp_fr, dep_fr_ptr *chain_dep_fr, or_fr_ptr *chain_or_fr);
#endif /* YAPOR */


/* ----------------
   Global functions
   ---------------- */

void private_completion(void) {
  sg_fr_ptr top_sg_fr;
  dep_fr_ptr dep_fr;

  optyap_message_type2("private completion --> completing all (sg_fr %x)",
                       (unsigned int)GEN_CP(B)->gcp_sg_fr);
  /* mark as complete all subgoals in the SCC */
  top_sg_fr = SgFr_next(GEN_CP(B)->gcp_sg_fr);
  do {
    SgFr_completed(LOCAL_top_sg_fr) = 1;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
  } while (LOCAL_top_sg_fr != top_sg_fr);
  /* release dependency frames */
  while (B > NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr))) {
    dep_fr = DepFr_next(LOCAL_top_dep_fr);
    FREE_DEPENDENCY_FRAME(LOCAL_top_dep_fr);
    LOCAL_top_dep_fr = dep_fr;
  }
  /* readjust registers */
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return;
}


#ifdef YAPOR
void public_completion(void) {
  dep_fr_ptr chain_dep_fr, next_dep_fr;
  or_fr_ptr chain_or_fr, top_or_fr, bottom_or_fr, next_or_fr;

  /* complete suspended subgoals */
  while (LOCAL_top_cp >= NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr))) {
    SgFr_completed(LOCAL_top_sg_fr) = 1;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
  }
  /* chain dependency frames to release */
  chain_dep_fr = NULL;
  while (LOCAL_top_cp >= NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr))) {
    LOCK(DepFr_lock(LOCAL_top_dep_fr));
    next_dep_fr = DepFr_next(LOCAL_top_dep_fr);
    DepFr_next(LOCAL_top_dep_fr) = chain_dep_fr;
    chain_dep_fr = LOCAL_top_dep_fr;
    LOCAL_top_dep_fr = next_dep_fr;
  }
  /* chain or-frames to release */
  chain_or_fr = NULL;
  bottom_or_fr = LOCAL_top_or_fr;
  top_or_fr = LOCAL_top_consumer_cp->cp_or_fr;
  while (top_or_fr != bottom_or_fr) {
    susp_fr_ptr aux_susp_fr;
    LOCK_OR_FRAME(top_or_fr);
    next_or_fr = OrFr_next(top_or_fr);
    OrFr_next(top_or_fr) = chain_or_fr;
    chain_or_fr = top_or_fr;
    aux_susp_fr = OrFr_suspensions(top_or_fr);
    while (aux_susp_fr) {
      release_suspension_branch(aux_susp_fr, &chain_dep_fr, &chain_or_fr);
      aux_susp_fr = SuspFr_next(aux_susp_fr);
    }
    top_or_fr = next_or_fr;
  }
  /* release dependency frames */
  while (chain_dep_fr) {
    next_dep_fr = DepFr_next(chain_dep_fr);
    FREE_DEPENDENCY_FRAME(chain_dep_fr);
    chain_dep_fr = next_dep_fr;
  }
  /* release or frames */
  while (chain_or_fr) {
    next_or_fr = OrFr_next(chain_or_fr);
    FREE_OR_FRAME(chain_or_fr);
    chain_or_fr = next_or_fr;
  }
  /* readjust registers */
  LOCAL_top_consumer_cp = NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr));
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return;
}


void suspend_completion_branch(void) {
  susp_fr_ptr new_susp_fr, aux_susp_fr;
  int h_size, b_size, tr_size;

  /* suspend current branch */
  h_size = H_FZ - B->cp_h;
  b_size = B - B_FZ;
  tr_size = TR_FZ - B->cp_tr;
  new_suspension_frame(new_susp_fr, worker_id, LOCAL_top_cp, 
                       LOCAL_top_consumer_cp, LOCAL_top_dep_fr, LOCAL_top_sg_fr, 
                       B->cp_h, B_FZ, B->cp_tr, h_size, b_size, tr_size);
  memcpy(SuspFr_global_start(new_susp_fr),
         SuspFr_global_reg(new_susp_fr),
         h_size);
  memcpy(SuspFr_local_start(new_susp_fr),
         SuspFr_local_reg(new_susp_fr),
         b_size);
  memcpy(SuspFr_trail_start(new_susp_fr),
         SuspFr_trail_reg(new_susp_fr),
         tr_size);
  /* store suspension frame in top or-frame */
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  aux_susp_fr = OrFr_suspensions(LOCAL_top_or_fr);
  SuspFr_next(new_susp_fr) = aux_susp_fr;
  OrFr_suspensions(LOCAL_top_or_fr) = new_susp_fr;
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);
  /* readjust registers */
  while (B >= NORM_CP(SgFr_gen_cp(LOCAL_top_sg_fr)))
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
  while (B > NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr)))
    LOCAL_top_dep_fr = DepFr_next(LOCAL_top_dep_fr);
  LOCAL_top_consumer_cp = NORM_CP(DepFr_cons_cp(LOCAL_top_dep_fr));
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return;
}


void resume_bottom_collected_suspension_branch(void) {
  susp_fr_ptr *susp_ptr, *bottom_susp_ptr, bottom_susp_fr;

  /* find bottom collected suspension frame */
  bottom_susp_ptr = & LOCAL_top_collect_susp_fr;
  susp_ptr = & SuspFr_next(LOCAL_top_collect_susp_fr);
  while (*susp_ptr) {
    if (SuspFr_top_cp(*bottom_susp_ptr) > SuspFr_top_cp(*susp_ptr))
      bottom_susp_ptr = susp_ptr;
    susp_ptr = & SuspFr_next(*susp_ptr);
  }
  bottom_susp_fr = *bottom_susp_ptr;
  *bottom_susp_ptr = SuspFr_next(*bottom_susp_ptr);
  /* resume suspension branch */
  memcpy(SuspFr_global_reg(bottom_susp_fr),
         SuspFr_global_start(bottom_susp_fr),
         SuspFr_global_size(bottom_susp_fr));
  memcpy(SuspFr_local_reg(bottom_susp_fr),
         SuspFr_local_start(bottom_susp_fr),
         SuspFr_local_size(bottom_susp_fr));
  memcpy(SuspFr_trail_reg(bottom_susp_fr),
         SuspFr_trail_start(bottom_susp_fr),
         SuspFr_trail_size(bottom_susp_fr));
  /* update shared nodes */
  /*
  SuspFr_susp_worker(bottom_susp_fr);
  */
  /* readjust registers */
  LOCAL_top_cp = SuspFr_top_cp(bottom_susp_fr);
  LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
  LOCAL_top_consumer_cp = SuspFr_top_consumer_cp(bottom_susp_fr);
  LOCAL_top_dep_fr = SuspFr_top_dep_fr(bottom_susp_fr);
  LOCAL_top_sg_fr = SuspFr_top_sg_fr(bottom_susp_fr);
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return;
}


void collect_suspension_frames_with_unconsumed_answers(void) {
  susp_fr_ptr *susp_ptr;
  susp_fr_ptr collect_susp_fr;

  susp_ptr = & OrFr_suspensions(LOCAL_top_or_fr);
  while (*susp_ptr) {
    collect_susp_fr = *susp_ptr;
    /* check for unconsumed answers in suspension frame */
    if (suspension_frame_with_unconsumed_answers(collect_susp_fr)) {
      /* collect suspension frame */
      *susp_ptr = SuspFr_next(collect_susp_fr);
      SuspFr_next(collect_susp_fr) = LOCAL_top_collect_susp_fr;
      LOCAL_top_collect_susp_fr = collect_susp_fr;
    }
    susp_ptr = & SuspFr_next(*susp_ptr);
  }
  return;
}


int suspension_frame_with_unconsumed_answers(susp_fr_ptr susp_fr) {
  dep_fr_ptr dep_fr;

  dep_fr = SuspFr_top_dep_fr(susp_fr);
  while (B > NORM_CP(DepFr_cons_cp(dep_fr))) {
    if (DepFr_last_ans(dep_fr) != SgFr_last_answer(DepFr_sg_fr(dep_fr)))
      return TRUE;
    dep_fr = DepFr_next(dep_fr);
  }
  return FALSE;
}


/* ---------------
   Local functions
   --------------- */

void release_suspension_branch(susp_fr_ptr susp_fr, dep_fr_ptr *chain_dep_fr, or_fr_ptr *chain_or_fr) {
  sg_fr_ptr aux_sg_fr;
  dep_fr_ptr aux_dep_fr, next_dep_fr;
  or_fr_ptr top_or_fr, bottom_or_fr, next_or_fr;

  /* complete suspended subgoals */
  aux_sg_fr = SuspFr_top_sg_fr(susp_fr);
  while (! SgFr_completed(aux_sg_fr) && SuspFr_top_cp(susp_fr) > NORM_CP(SgFr_gen_cp(aux_sg_fr))) {
    SgFr_completed(aux_sg_fr) = 1;
    aux_sg_fr = SgFr_next(aux_sg_fr);
  }
  /* chain dependency frames to release */
  aux_dep_fr = SuspFr_top_dep_fr(susp_fr);
  while (! IS_LOCKED(DepFr_lock(aux_dep_fr)) && SuspFr_top_cp(susp_fr) > NORM_CP(DepFr_cons_cp(aux_dep_fr))) {
    LOCK(DepFr_lock(aux_dep_fr));
    next_dep_fr = DepFr_next(aux_dep_fr);
    DepFr_next(aux_dep_fr) = *chain_dep_fr;
    *chain_dep_fr = aux_dep_fr;
    aux_dep_fr = next_dep_fr;
  }
  /* chain or-frames to release */
  bottom_or_fr = SuspFr_top_cp(susp_fr)->cp_or_fr;
  top_or_fr = SuspFr_top_consumer_cp(susp_fr)->cp_or_fr;
  while (top_or_fr != bottom_or_fr) {
    susp_fr_ptr aux_susp_fr;
    LOCK_OR_FRAME(top_or_fr);
    next_or_fr = OrFr_next(top_or_fr);
    OrFr_next(top_or_fr) = *chain_or_fr;
    *chain_or_fr = top_or_fr;
    aux_susp_fr = OrFr_suspensions(top_or_fr);
    while (aux_susp_fr) {
      release_suspension_branch(aux_susp_fr, chain_dep_fr, chain_or_fr);
      aux_susp_fr = SuspFr_next(aux_susp_fr);
    }
    top_or_fr = next_or_fr;
  }
  return;
}
#endif /* YAPOR */
#endif /* TABLING */
