/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		adtdefs.c						 *
* Last rev:								 *
* mods:									 *
* comments:	abstract machine definitions				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char SccsId[] = "%W% %G%";

#endif

#define ADTDEFS_C

#include "Yap.h"
#include "Yatom.h"
#include <stdio.h>

static Prop
GetFunctorProp(Atom a, unsigned int arity)
{				/* look property list of atom a for kind  */
  FunctorEntry *pp = RepFunctorProp(RepAtom(a)->PropOfAE);

  while (!EndOfPAEntr(pp) &&
	 (pp->KindOfPE != FunctorProp ||
	  pp->ArityOfFE != arity))
    pp = RepFunctorProp(pp->NextOfPE);
  return (AbsFunctorProp(pp));
}

/* vsc: We must guarantee that IsVarTerm(functor) returns true! */
Functor
MkFunctor(Atom ap, unsigned int arity)
{

  FunctorEntry *p;
  Prop p0 = GetFunctorProp(ap, arity);

  if (p0)
    return ((Functor) RepProp(p0));
  YAPEnterCriticalSection();
  p = (FunctorEntry *) AllocAtomSpace(sizeof(*p));
  p->KindOfPE = FunctorProp;
  p->NameOfFE = ap;
  p->ArityOfFE = arity;
  p->NextOfPE = RepAtom(ap)->PropOfAE;
  RepAtom(ap)->PropOfAE = AbsProp((PropEntry *) p);
  YAPLeaveCriticalSection();
  return ((Functor) p);
}

#define HashFunction(c,h) h=0; while(*c) h += *c++; h%=MaxHash

inline static Atom
SearchInInvisible(char *atom)
{
  AtomEntry *chain = RepAtom(*INVISIBLECHAIN);

  while (!EndOfPAEntr(chain) && strcmp(chain->StrOfAE, atom) != 0)
    chain = RepAtom(chain->NextOfAE);
  if (chain == NIL)
    return (0);
  return (AbsAtom(chain));
}

Atom
LookupAtom(char *atom)
{				/* lookup atom in atom table            */
  register CELL hash;
  register char *p;
  Atom a;
  AtomEntry *ae;

  if (LookupMode == FullLookup) {
    Atom t;

    if ((t = SearchInInvisible(atom)) != 0)
      return (t);
  }
  /* compute hash */
  p = atom;
  HashFunction(p, hash);
  a = HashChain[hash];
  /* search atom in chain */
  while (a != NIL) {
    ae = RepAtom(a);
    if (strcmp(ae->StrOfAE, atom) == 0)
      return (a);
    a = ae->NextOfAE;
  }
  /* add new atom to start of chain */
  YAPEnterCriticalSection();
  ae = (AtomEntry *) AllocAtomSpace((sizeof *ae) + strlen(atom));
  a = AbsAtom(ae);
  ae->NextOfAE = HashChain[hash];
  HashChain[hash] = a;
  ae->PropOfAE = NIL;
  if (ae->StrOfAE != atom)
    strcpy(ae->StrOfAE, atom);
  YAPLeaveCriticalSection();
  return (a);
}

void
ReleaseAtom(Atom atom)
{				/* Releases an atom from the hash chain */
  register Int hash;
  register char *p;
  AtomEntry *inChain;
  char *name = (RepAtom(atom))->StrOfAE;

  /* compute hash */
  p = name;
  HashFunction(p, hash);
  /* if the first in the chain */
  if (HashChain[hash] == atom) {
    YAPEnterCriticalSection();
    HashChain[hash] = (RepAtom(atom))->NextOfAE;
    YAPLeaveCriticalSection();
    return;
  }
  /* else */
  inChain = RepAtom(HashChain[hash]);
  while (inChain->NextOfAE != atom)
    inChain = RepAtom(inChain->NextOfAE);
  YAPEnterCriticalSection();
  inChain->NextOfAE = (RepAtom(atom))->NextOfAE;
  YAPLeaveCriticalSection();
}

Prop
GetAProp(Atom a, PropFlags kind)
{				/* look property list of atom a for kind  */
  PropEntry *pp = RepProp((RepAtom(a))->PropOfAE);

  while (!EndOfPAEntr(pp) && pp->KindOfPE != kind)
    pp = RepProp(pp->NextOfPE);
  return (AbsProp(pp));
}

Prop
GetPredProp(Atom ap, unsigned int arity)
     /* get predicate entry for ap/arity;               */
{
  Prop p0;
  PredEntry *p = RepPredProp(p0 = RepAtom(ap)->PropOfAE);

  while (p0 && (p->KindOfPE != PEProp || p->ArityOfPE != arity ||
		(p->ModuleOfPred && p->ModuleOfPred != CurrentModule)))
    p = RepPredProp(p0 = p->NextOfPE);
  return (p0);
}

Prop
PredProp(Atom ap, unsigned int arity)
     /* get predicate entry for ap/arity; create it if neccessary.              */
{
  Prop p0;
  PredEntry *p = RepPredProp(p0 = RepAtom(ap)->PropOfAE);

  while (p0 && (p->KindOfPE != 0 || p->ArityOfPE != arity ||
		(p->ModuleOfPred && p->ModuleOfPred != CurrentModule)))
    p = RepPredProp(p0 = p->NextOfPE);

  if (p0 != NIL)
    return (p0);
  YAPEnterCriticalSection();
  p = (PredEntry *) AllocAtomSpace(sizeof(*p));
#ifdef YAPOR
  INIT_LOCK(p->lock);
  INIT_LOCK(p->indexado);
#endif	/* YAPOR */
  p->KindOfPE = PEProp;
  p->ArityOfPE = arity;
  p->FirstClause = p->LastClause = NIL;
  p->PredFlags = 0L;
  p->StateOfPred = 0;
  p->Owner.OwnerFile = AtomNil;
  p->OpcodeOfPred = opcode(_undef_p);
  p->TrueCodeOfPred = p->CodeOfPred = (CODEADDR)(&(p->OpcodeOfPred)); 
  p->ModuleOfPred = CurrentModule;
  if (arity == 0)
    p->FunctorOfPred = (Functor) ap;
  else
    p->FunctorOfPred = MkFunctor(ap, arity);
  p->NextOfPE = RepAtom(ap)->PropOfAE;
#ifdef TABLING
  p->TableOfPred = NULL;
#endif /* TABLING */
  RepAtom(ap)->PropOfAE = p0 = AbsPredProp(p);
  YAPLeaveCriticalSection();
  return (p0);
}

Term
GetValue(Atom a)
{
  Prop p0 = GetAProp(a, ValProperty);

  if (p0 == 0)
    return (TermNil);
  return (RepValProp(p0)->ValueOfVE);
}

void
PutValue(Atom a, Term v)
{
  Prop p0 = GetAProp(a, ValProperty);
  ValEntry *p;

  YAPEnterCriticalSection();
  if (p0 != NIL) {
    p = RepValProp(p0);
  } else {
    p = (ValEntry *) AllocAtomSpace(sizeof(ValEntry));
    p->NextOfPE = RepAtom(a)->PropOfAE;
    RepAtom(a)->PropOfAE = AbsValProp(p);
    p->KindOfPE = ValProperty;
  }
  if (IsFloatTerm(v)) {
    /* store a float in code space, so that we can access the property */
    union {
      Float f;
      CELL ar[sizeof(Float) / sizeof(CELL)];
    } un;
    CELL *pt;
    unsigned int i;

    un.f = FloatOfTerm(v);
    if (p0 != NIL && IsApplTerm(p->ValueOfVE))
      pt = RepAppl(p->ValueOfVE);
    else {
      pt = (CELL *) AllocAtomSpace(2 * sizeof(CELL) + sizeof(Float));
    }

    pt[0] = (CELL)FunctorDouble;
    pt[1 + sizeof(Float) / sizeof(CELL)] = (CELL) FunctorEndDouble;
    for (i = 0; i < sizeof(Float) / sizeof(CELL); i++)
      pt[i + 1] = un.ar[i];
    p->ValueOfVE = AbsAppl(pt);
  } else if (IsLongIntTerm(v)) {
    CELL *pt;
    if (p0 != NIL && IsApplTerm(p->ValueOfVE)) { 
      pt = RepAppl(p->ValueOfVE);
    } else {
      pt = (CELL *) AllocAtomSpace(2 * sizeof(CELL) + sizeof(Float));
    }
    pt[0] = (CELL)FunctorLongInt;
    pt[1] = (CELL)LongIntOfTerm(v);
    pt[2] = (CELL)FunctorEndLongInt;
    p->ValueOfVE = AbsAppl(pt);
  } else {
    if (p0 != NIL && IsApplTerm(p->ValueOfVE)) {
      /* recover space */
      FreeCodeSpace((char *) (RepAppl(p->ValueOfVE)));
    }
    p->ValueOfVE = v;
  }
  YAPLeaveCriticalSection();
}

/* compare a float with a possibly unaligned float in memory */
int 
compare_float_mem(Float f1, CELL *f2)
{
  union {
    Float f;
    CELL ar[sizeof(Float) / sizeof(CELL)];
  } un;
  unsigned int i;

  un.f = f1;
  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    if (un.ar[i] != f2[i])
      return (FALSE);
  }
  return (TRUE);
}

Term
MkFloatFromCells(CELL *ptr)
{
  unsigned int i;
  register CELL *iptr = H;
  register CELL out;

  out = AbsAppl(iptr);
  *iptr++ = (CELL) FunctorDouble;
  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    *iptr++ = *ptr++;
  }
  *iptr = (CELL) FunctorEndDouble;
  H = iptr + 1;
  return (out);
}

Float
FloatOfTerm(Term t)
{

  CELL *iptr = RepAppl(t) + 1;
  union {
    Float f;
    CELL ar[sizeof(Float) / sizeof(CELL)];
  } un;
  unsigned int i;

#ifdef DEBUG
  if (iptr[-1] != (CELL) FunctorDouble) {
    Error(SYSTEM_ERROR, t, "corrupted float");
    return(0.0);
  }
#endif	/* DEBUG */
  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    un.ar[i] = *iptr++;
  }

  return (un.f);
}

Term
MkFloatTerm(Float flt)
{
  register CELL *iptr = H;
  union {
    Float f;
    CELL ar[sizeof(Float) / sizeof(CELL)];
  } un;
  CELL out = AbsAppl(iptr);
  unsigned int i;

  un.f = flt;

  *iptr++ = (CELL) FunctorDouble;
  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    *iptr++ = un.ar[i];
  }
  *iptr++ = (CELL) FunctorEndDouble;
  H = iptr;
  return (out);

}

Term
StringToList(register char *s)
{
  register Term t;
  register char *cp = s + strlen(s);

  t = MkAtomTerm(AtomNil);
  while (cp > s) {
    t = MkPairTerm(MkIntTerm(*--cp), t);
  }
  return (t);
}

Term
StringToListOfAtoms(register char *s)
{
  register Term t;
  char so[2];
  register char *cp = s + strlen(s);

  so[1] = '\0';
  t = MkAtomTerm(AtomNil);
  while (cp > s) {
    so[0] = *--cp;
    t = MkPairTerm(MkAtomTerm(LookupAtom(so)), t);
  }
  return (t);
}

Term
ArrayToList(register Term *tp, int nof)
{
  register Term *pt = tp + nof;
  register Term t;

  t = MkAtomTerm(AtomNil);
  while (pt > tp) {
    Term tm = *--pt;
#if SBA
    if (tm == 0)
      t = MkPairTerm((CELL)pt, t);
    else
#endif
      t = MkPairTerm(tm, t);
  }
  return (t);
}

int
GetName(char *s, Term t)
{
  register Term Head;
  register Int i;

  if (IsVarTerm(t) || !IsPairTerm(t))
    return (FALSE);
  while (IsPairTerm(t)) {
    Head = HeadOfTerm(t);
    if (!IsNumTerm(Head))
      return (FALSE);
    i = IntOfTerm(Head);
    if (i < 0 || i > 255)
      return (FALSE);
    *s++ = i;
    t = TailOfTerm(t);
  }
  *s = '\0';
  return (TRUE);
}

#ifdef SFUNC

Term
MkSFTerm(Functor f, int n, Term *a, empty_value)
{
  Term t, p = AbsAppl(H);
  int i;

  *H++ = f;
  RESET_VARIABLE(H);
  ++H;
  for (i = 1; i <= n; ++i) {
    t = Derefa(a++);
    if (t != empty_value) {
      *H++ = i;
      *H++ = t;
    }
  }
  *H++ = 0;
  return (p);
}

CELL *
ArgsOfSFTerm(Term t)
{
  CELL *p = RepAppl(t) + 1;

  while (*p != (CELL) p)
    p = CellPtr(*p) + 1;
  return (p + 1);
}

#endif
