/*************************************************************************
*									 *
*	 YAP Prolog  %W% %G%				*
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		alloc.h							 *
* Last rev:								 *
* mods:									 *
* comments:	allocating space					 *
*									 *
*************************************************************************/

/* Space Organization:
	The data areas is divided in the following way:

 Lower Addresses
	HeapBase	\	
				 |	Atom and code space
	HeapTop-1	/
	
	HeapTop		\
				 | 	Free Space
	AuxSp		\
				 |	Auxiliary stack
	AuxTop		/

	GlobalBase-1/

	GlobalBase	\	
				 |	Global Stack
	H			/
	
	ASP			\
				 |	Local stack
	LocalBase	/

	TRBase			/
	
	TR		\
				 |  Trail
 Higher Adresses

*/

/* definitions required by saver/restorer and memory manager */

typedef	CELL SEG_SIZE;

typedef	struct FREEB {
	SEG_SIZE		b_size;
	struct FREEB		*b_next;
	struct FREEB		*b_next_size;
	} BlockHeader;

#define MinBlockSize	(sizeof(BlockHeader)+sizeof(SEG_SIZE))
#define MaxBlockSize	0xffffff
#define	InUseFlag	0x1000000

/* the following defines are machine dependant and are used to enforce
   the correct alignment for allocated blocks							*/

#if SIZEOF_INT_P==4
#define ALIGN			3
#define ALIGNMASK		0xfffffffc
#else
#define ALIGN			7
#define ALIGNMASK		0xfffffff8L
#endif	/* ALIGN_LONGS */

#define AdjustSize(X)	((X+ALIGN) & ALIGNMASK)


#define BlockTrailer(b)		((SEG_SIZE *)b)[((BlockHeader *) b)->b_size]

#ifndef YAPOR
extern BlockHeader *FreeBlocks;
#endif /* YAPOR */

/* Operating system and architecture dependent page size */
extern int page_size;

#if USE_MMAP


#elif USE_SHM



#elif USE_SBRK

#ifdef __svr4__
#include <unistd.h>
#elif sun
#include <sys/types.h>
#include <malloc.h>
void *sbrk(caddr_t);

typedef unsigned size_t;

MALLOC_T malloc(size_t);
void free(MALLOC_T);
MALLOC_T realloc(MALLOC_T,size_t);
MALLOC_T calloc(size_t,size_t);

#endif


#else


#endif












