/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		evalis.c						 *
* Last rev:								 *
* mods:									 *
* comments:	is/3 predicate						 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif /* SCCS */

/*
 * This predicates had to be developed here because of a bug in the MPW
 * compiler, which was not able to compile the original eval.c 
 */

#include "Yap.h"
#include "Yatom.h"
#include "eval.h"

#define IntRes(X)       return(unify_constant(ARG1,MkIntegerTerm(X)))
#define FloatRes(X)	return(unify_constant(ARG1,MkEvalFl(X)))

Int 
p_unary_is()
{
  register SMALLUNSGN OpNum;
  Term t2, t3;
  Int iop1;
  Float fop1;

  t2 = Deref(ARG2);
  if (IsAtomTerm(t2)) {
    Atom            name;
    Prop            p;
    name = AtomOfTerm(t2);
    if ((p = GetAProp(name, ExpProperty(1))) == NIL) {
      Term t, ti[2];

      ti[0] = MkAtomTerm(name);
      ti[1] = MkIntegerTerm(1);
      t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
      Error(TYPE_ERROR_EVALUABLE, t,
	    "arithmetic expression %s/%d",
	    RepAtom(name)->StrOfAE,
	    1
	    );
      P = (yamop *)FAILCODE;
      return(FALSE);
    }
    OpNum = RepExpProp(p)->ENoOfEE;
  } else if (IsIntTerm(t2))
    OpNum = IntOfTerm(t2);
  else
    return (FALSE);
  t3 = Deref(ARG3);
  if (IsVarTerm(t3)) {
    int op = 0;
    
    while (InitTab[op].eno != OpNum) op++;
    Error(INSTANTIATION_ERROR, t3, "arithmetic expression %s/1", InitTab[op].OpName);
    P = (yamop *)FAILCODE;
    return(FALSE);
  }
  if (IsIntTerm(t3)) {
    iop1 = IntOfTerm(t3);

  doint:
    switch (OpNum) {
    case e_uminus:
      IntRes(-iop1);
    case e_abs:
#if SHORT_INTS
#if HAVE_LABS
      IntRes((Int)labs((long int)iop1));
#else
      IntRes((Iop1 >= 0 ? iop1 : -iop1));
#endif
#else
      IntRes(abs(iop1));
#endif
    case e_msb:
      IntRes(msb(iop1));
    case e_uplus:
      IntRes(iop1);
    case e_not:
      IntRes(~iop1);
    case e_exp:
      FloatRes(exp(FL(iop1)));
    case e_log:
      FloatRes(log(FL(iop1)));
    case e_log10:
      FloatRes(log10(FL(iop1)));
    case e_sqrt:
      FloatRes(sqrt(FL(iop1)));
    case e_sin:
      FloatRes(sin(FL(iop1)));
    case e_cos:
      FloatRes(cos(FL(iop1)));
    case e_tan:
      FloatRes(tan(FL(iop1)));
    case e_sinh:
      FloatRes(sinh(FL(iop1)));
    case e_cosh:
      FloatRes(cosh(FL(iop1)));
    case e_tanh:
      FloatRes(tanh(FL(iop1)));
    case e_asin:
      FloatRes(asin(FL(iop1)));
    case e_acos:
      FloatRes(acos(FL(iop1)));
    case e_atan:
      FloatRes(atan(FL(iop1)));
    case e_asinh:
      FloatRes(asinh(FL(iop1)));
    case e_acosh:
      FloatRes(acosh(FL(iop1)));
    case e_atanh:
      FloatRes(atanh(FL(iop1)));
    case e_floor:
    case e_ceiling:
    case e_round:
    case e_truncate:
      FloatRes(FL(iop1));
    case e_integer:
      IntRes(iop1);
    case e_float:
      FloatRes(FL(iop1));
    case e_fmodf:
      FloatRes(FL(0.0));
    case e_imodf:
      FloatRes(FL(Iop1));
    default:
      {
	Term t, ti[2];
	
	ti[0] = MkIntTerm(OpNum);
	ti[1] = MkIntegerTerm(1);
	t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	Error(TYPE_ERROR_EVALUABLE, t,
	      "arithmetic expression %d/%d",
	      OpNum,
	      1
	      );
      }
      P = (yamop *)FAILCODE;
      return(FALSE);
    }
  } else if (IsFloatTerm(t3)) {
    fop1 = FloatOfTerm(t3);
  dofloat:
    switch (OpNum) {
    case e_uminus:
      FloatRes(-fop1);
    case e_abs:
      FloatRes(fabs(fop1));
    case e_uplus:
      FloatRes(fop1);
    case e_exp:
      FloatRes(exp(fop1));
    case e_log:
      FloatRes(log(fop1));
    case e_log10:
      FloatRes(log10(fop1));
    case e_sqrt:
      FloatRes(sqrt(fop1));
    case e_sin:
      FloatRes(sin(fop1));
    case e_cos:
      FloatRes(cos(fop1));
    case e_tan:
      FloatRes(tan(fop1));
    case e_sinh:
      FloatRes(sinh(fop1));
    case e_cosh:
      FloatRes(cosh(fop1));
    case e_tanh:
      FloatRes(tanh(fop1));
    case e_asin:
      FloatRes(asin(fop1));
    case e_acos:
      FloatRes(acos(fop1));
    case e_atan:
      FloatRes(atan(fop1));
    case e_asinh:
      FloatRes(asinh(fop1));
    case e_acosh:
      FloatRes(acosh(fop1));
    case e_atanh:
      FloatRes(atanh(fop1));
    case e_floor:
      FloatRes(floor(fop1));
    case e_ceiling:
      FloatRes(ceil(fop1));
    case e_round:
      FloatRes(rint(fop1));
    case e_truncate:
      FloatRes((fop1 >= 0.0 ? floor(fop1) : ceil(fop1)) );
    case e_integer:
      IntRes((Int) fop1);
      if (finite(fop1))
	IntRes((Int) fop1);
      else {
	Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fop1), "integer/1");
	P = (yamop *)FAILCODE;
	return(FALSE);	    
      }
    case e_float:
      FloatRes(fop1);
    case e_fmodf:
      {
	Float dbl;
	FloatRes(modf(Fop1,&dbl));
      }
    case e_imodf:
      {
	Float dbl;

	modf(Fop1,&dbl);
	FloatRes(dbl);
      }
    case e_msb:
      Error(TYPE_ERROR_INTEGER, t3, "msb/1");
      P = (yamop *)FAILCODE;
      return(FALSE);
    case e_not:
      Error(TYPE_ERROR_INTEGER, t3, "not/1");
      P = (yamop *)FAILCODE;
      return(FALSE);
    default:
      {
	Term t, ti[2];

	ti[0] = MkIntTerm(OpNum);
	ti[1] = MkIntegerTerm(2);
	t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	Error(TYPE_ERROR_EVALUABLE, t,
	      "arithmetic expression %d/%d",
	      OpNum,
	      2
	      );
      }
      P = (yamop *)FAILCODE;
      return(FALSE);
    }
  } else if (IsLongIntTerm(t3)) {
    iop1 = LongIntOfTerm(t3);
    goto doint;
  } else if (Eval(t3) == FInt) {
    iop1 = Iop1;
    goto doint;
  } else {
    fop1 = Fop1;
    goto dofloat;
  }
}

