/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		evaltwo.c						 *
* Last rev:								 *
* mods:									 *
* comments:	is/4 predicate						 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif 

/*
 * This predicates had to be developed here because of a bug in the MPW
 * compiler, which was not able to compile the original eval.c 
 */

#include "Yap.h"
#include "Yatom.h"
#include "eval.h"

#define IntRes(X)       return(unify_constant(ARG1,MkIntegerTerm(X)))
#define FloatRes(X)	return(unify_constant(ARG1,MkEvalFl(X)))

Int
p_binary_is()
{
  register SMALLUNSGN OpNum;
  Term t2,t3;
  Int iop1;
  Float fop1;

  t2 = Deref(ARG2);
  if (IsIntTerm(t2))
    OpNum = IntOfTerm(t2);
  else if (IsAtomTerm(t2)) {
    Atom            name = AtomOfTerm(t2);
    Prop            p;
    if ((p = GetAProp(name, ExpProperty(2))) == NIL) {
      Term t, ti[2];

      ti[0] = MkIntegerTerm(2);
      ti[0] = MkAtomTerm(name);
      t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
      Error(TYPE_ERROR_EVALUABLE, t,
	    "arithmetic expression %s/%d",
	    RepAtom(name)->StrOfAE,
	    2
	    );
      P = (yamop *)FAILCODE;
      return(FALSE);
    }
    OpNum = RepExpProp(p)->ENoOfEE;
  } else
    return (FALSE);
  t3 = Deref(ARG3);
  if (IsVarTerm(t3)) {
    int op = 0;
    
    while (InitTab[op].eno != OpNum) op++;
    Error(INSTANTIATION_ERROR, t3,
	  "arithmetic expression %s/2", InitTab[op].OpName);
    return(FALSE);
  } else if (IsIntTerm(t3)) {
    int iop2;
    Float fop2;
    Term t4;

    iop1 = IntOfTerm(t3);
    /* case for first argument integer */
  int1:
    t4 = Deref(ARG4);
    if (IsVarTerm(t4)) {
      int op = 0;
    
      while (InitTab[op].eno != OpNum) op++;
      Error(INSTANTIATION_ERROR, t4,
	    "arithmetic expression %s/2", InitTab[op].OpName);
      return(FALSE);
    } else if (IsIntTerm(t4)) {
      iop2 = IntOfTerm(t4);
    ii:
      /* both arguments are integers */
      switch (OpNum) {
      case e_div:
#ifdef TRY_TO_CONVERT_FLOATS_TO_INTS
	if (iop1 % iop2 == 0)
	  IntRes(iop1 / iop2);
#endif
	FloatRes(FL(iop1) / FL(iop2));
      case e_and:
	IntRes(iop1 & iop2);
      case e_or:
	IntRes(iop1 | iop2);
      case e_lshift:
	IntRes(iop1 << iop2);
      case e_rshift:
	IntRes(iop1 >> iop2);
      case e_power:
	FloatRes(pow(FL(iop1), FL(iop2)));
      case e_atan2:
	FloatRes(atan2(FL(iop1), FL(iop2)));
      case e_plus:
	IntRes(iop1 + iop2);
      case e_dif:
	IntRes(iop1 - iop2);
      case e_times:
	IntRes(iop1 * iop2);
      case e_idiv:
	if (iop2 == 0) {
	  Error(EVALUATION_ERROR_ZERO_DIVISOR,TermNil,"// /2");
	  P = (yamop *)FAILCODE;
	  IntRes(0);
	}
	IntRes(iop1 / iop2);
      case e_mod:
	IntRes(iop1 % iop2);
      case e_gcd:
	IntRes(gcd(abs(iop1),abs(iop2)));
      case e_gcdmult:
	{
	  Int x;
	  IntRes(gcdmult(abs(iop1),abs(iop2), &x));
	}
      case e_min:
	IntRes((iop1 < iop2 ? iop1 : iop2));
      case e_max:
	IntRes((iop1 > iop2 ? iop1 : iop2));
      default:
	{
	  Term t, ti[2];
    
	  ti[0] = MkIntTerm(OpNum);
	  ti[1] = MkIntegerTerm(2);
	  t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	  Error(TYPE_ERROR_EVALUABLE, t4,
		"arithmetic expression %d/%d",
		OpNum,
		2
		);
	  P = (yamop *)FAILCODE;
	}
	return(FALSE);
      }
    } else if (IsFloatTerm(t4)) {
      fop2 = FloatOfTerm(t4);
    intf:
      /* integer, float */
      switch (OpNum) {
      case e_power:
	FloatRes(pow(FL(iop1), fop2));
      case e_atan2:
	FloatRes(atan2(FL(iop1), fop2));
      case e_plus:
	FloatRes(iop1+fop2);
      case e_dif:
	FloatRes(iop1 - fop2);
      case e_times:
	FloatRes(iop1 * fop2);
      case e_div:
	FloatRes(iop1 / fop2);
      case e_min:
	if (iop1 <= fop2) {
	  IntRes(iop1);
	} else {
	  FloatRes(fop2);
	}
      case e_max:
	if (iop1 >= fop2) {
	  IntRes(iop1);
	} else {
	  FloatRes(fop2);
	}
      case e_lshift:
	Error(TYPE_ERROR_INTEGER, t4, "<</2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_rshift:
	Error(TYPE_ERROR_INTEGER, t4, ">>/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_and:
	Error(TYPE_ERROR_INTEGER, t4, "/\\/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_or:
	Error(TYPE_ERROR_INTEGER, t4, "\\//2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_mod:
	Error(TYPE_ERROR_INTEGER, t4, "mod/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_idiv:
	Error(TYPE_ERROR_INTEGER, t4, "// /2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcd:
	Error(TYPE_ERROR_INTEGER, t4, "gcd/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcdmult:
	Error(TYPE_ERROR_INTEGER, t4, "gcdmult/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      default:
	{
	  Term t, ti[2];
    
	  ti[0] = MkIntTerm(OpNum);
	  ti[1] = MkIntegerTerm(2);
	  t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	  Error(TYPE_ERROR_EVALUABLE, t,
		"arithmetic expression %d/%d",
		OpNum,
		2
		);
	  P = (yamop *)FAILCODE;
	}
	return(FALSE);
      }
    } else if (IsLongIntTerm(t4)) {
      iop2 = LongIntOfTerm(t4);
      goto ii;
    } else if (Eval(t4) == FInt) {
      iop2 = Iop1;
      goto ii;
    } else {
      fop2 = Fop1;
      goto intf;
    }
  } else if (IsFloatTerm(t3)) {
    Int iop2;
    Float fop2;
    Term t4;

    fop1 = FloatOfTerm(t3);
    /* case for first argument float */
  fl1:
    t4 = Deref(ARG4);
    if (IsVarTerm(t4)) {
      int op = 0;
    
      while (InitTab[op].eno != OpNum) op++;
      Error(INSTANTIATION_ERROR, t4,
	  "arithmetic expression %s/2", InitTab[op].OpName);
      P = (yamop *)FAILCODE;
      return(FALSE);
    } else if (IsIntTerm(t4)) {
      iop2 = IntOfTerm(t4);
    fi:
      /* float, integer */
      switch (OpNum) {
      case e_power:
	FloatRes(pow(fop1, FL(iop2)));
      case e_atan2:
	FloatRes(atan2(fop1, FL(iop2)));
      case e_plus:
	FloatRes(fop1+iop2);
      case e_dif:
	FloatRes(fop1 - iop2);
      case e_times:
	FloatRes(fop1 * iop2);
      case e_div:
	FloatRes(fop1 / iop2);
      case e_min:
	if (fop1 < iop2) {
	  FloatRes(fop1);
	} else {
	  IntRes(iop2);
	}
      case e_max:
	if (fop1 > iop2) {
	  FloatRes(fop1);
	} else {
	  IntRes(iop2);
	}
      case e_lshift:
	Error(TYPE_ERROR_INTEGER, t3, "<</2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_rshift:
	Error(TYPE_ERROR_INTEGER, t3, ">>/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_and:
	Error(TYPE_ERROR_INTEGER, t3, "/\\/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_or:
	Error(TYPE_ERROR_INTEGER, t3, "\\//2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_mod:
	Error(TYPE_ERROR_INTEGER, t3, "mod/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_idiv:
	Error(TYPE_ERROR_INTEGER, t3, "// /2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcd:
	Error(TYPE_ERROR_INTEGER, t3, "gcd/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcdmult:
	Error(TYPE_ERROR_INTEGER, t3, "gcdmult/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      default:
	{
	  Term t, ti[2];
    
	  ti[0] = MkIntTerm(OpNum);
	  ti[1] = MkIntegerTerm(2);
	  t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	  Error(TYPE_ERROR_EVALUABLE, t,
		"arithmetic expression %d/%d",
		OpNum,
		2
		);
	  P = (yamop *)FAILCODE;
	}
	return(FALSE);
      }
    } else if (IsFloatTerm(t4)) {
      fop2 = FloatOfTerm(t4);
    ff:
      /* integer, float */
      switch (OpNum) {
      case e_power:
	FloatRes(pow(fop1, fop2));
      case e_atan2:
	FloatRes(atan2(fop1, fop2));
      case e_plus:
	FloatRes(fop1+fop2);
      case e_dif:
	FloatRes(fop1 - fop2);
      case e_times:
	FloatRes(fop1 * fop2);
      case e_div:
	FloatRes(fop1 / fop2);
      case e_min:
	if (fop1 <= fop2) {
	  FloatRes(fop1);
	} else {
	  FloatRes(fop2);
	}
      case e_max:
	if (fop1 >= fop2) {
	  FloatRes(fop1);
	} else {
	  FloatRes(fop2);
	}
      case e_lshift:
	Error(TYPE_ERROR_INTEGER, t4, "<</2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_rshift:
	Error(TYPE_ERROR_INTEGER, t4, ">>/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_and:
	Error(TYPE_ERROR_INTEGER, t4, "/\\/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_or:
	Error(TYPE_ERROR_INTEGER, t4, "\\//2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_mod:
	Error(TYPE_ERROR_INTEGER, t4, "mod/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_idiv:
	Error(TYPE_ERROR_INTEGER, t4, "// /2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcd:
	Error(TYPE_ERROR_INTEGER, t4, "gcd/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      case e_gcdmult:
	Error(TYPE_ERROR_INTEGER, t4, "gcdmult/2");
	P = (yamop *)FAILCODE;
	return(FALSE);
      default:
	{
	  Term t, ti[2];
    
	  ti[0] = MkIntTerm(OpNum);
	  ti[1] = MkIntegerTerm(2);
	  t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
	  Error(TYPE_ERROR_EVALUABLE, t,
		"arithmetic expression %d/%d",
		OpNum,
		2
		);
	}
	P = (yamop *)FAILCODE;
	return(FALSE);
      }
    } else if (IsLongIntTerm(t4)) {
      iop2 = LongIntOfTerm(t4);
      goto fi;

    } else if (Eval(t4) == FInt) {
      iop2 = Iop1;
      goto fi;
    } else {
      fop2 = Fop1;
      goto ff;
    }
  } else if (IsLongIntTerm(t3)) {
    iop1 = LongIntOfTerm(t3);
    goto int1;
  } else if (Eval(t3) == FInt) {
    iop1 = Iop1;
    goto int1;
  } else {
    fop1 = Fop1;
    goto fl1;
  }
}
