/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		preds.yap						 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	Predicate Manipulation for YAP				 *
*									 *
*************************************************************************/

% The next predicates are applicable only
% to dynamic code

asserta(V) :- var(V), !,
	throw(error(instantiation_error,asserta(V))).
asserta(M:C) :- !, call(M:asserta(C)).
asserta(C) :- '$assert'(C,first,_,asserta(C)).

assertz(V) :- var(V), !,
	throw(error(instantiation_error,assertz(V))).
assertz(M:C) :- !, call(M:assertz(C)).
assertz(C) :- '$assert'(C,last,_,assertz(C)).

assert(V) :- var(V), !,
	throw(error(instantiation_error,assert(V))).
assert(M:C) :- !, call(M:assert(C)).
assert(C) :- '$assert'(C,last,_,assert(C)).

'$assert'(C0,Where,R,P) :-
    '$expand_term_modules'(C0,C),
    '$check_head_and_body'(C,H,B,P),
    ( '$is_dynamic'(H) -> '$getenv'('HOME',_), '$assertat_d'(Where,H,B,C,R)
            ;
      '$undefined'(H) -> 
		functor(H, Na, Ar),
		dynamic(Na/Ar),
		'$assertat_d'(Where,H,B,C,R)
	    ;
	'$assert1'(Where,C,H)
    ).

assert_static(V) :- var(V), !,
	throw(error(instantiation_error,assert_static(V))).
assert_static(M:C) :- !, call(M:assert_static(C)).
assert_static(C) :- '$assert_static'(C,last,_,assert_static(C)).

asserta_static(V) :- var(V), !,
	throw(error(instantiation_error,asserta_static(V))).
asserta_static(M:C) :- !, call(M:asserta_static(C)).
asserta_static(C) :- '$assert_static'(C,first,_,asserta_static(C)).

assertz_static(V) :- var(V), !,
	throw(error(instantiation_error,assertz_static(V))).
assertz_static(M:C) :- !, call(M:assertz_static(C)).
assertz_static(C) :- '$assert_static'(C,last,_,assertz_static(C)).

'$assert_static'(C0,Where,R,P) :-
    '$expand_term_modules'(C0,C),
    '$check_head_and_body'(C,H,B,P),
    ( '$is_dynamic'(H) -> throw(error(permission_error(modify,dynamic_procedure,Na/Ar),P)) ;
      '$undefined'(H), '$get_value'('$full_iso',true) ->  functor(H,Na,Ar), dynamic(Na/Ar), '$assertat_d'(Where,H,B,C,R) ;
	'$assert1'(Where,C,H)
    ).


'$assertat_d'(first,Head,Body,C0,R) :- !,
		'$compile_dynamic'((Head:-Body),2,CR),
		'$head_and_body'(C0, H0, B0),
		'$recordap'(Head,(H0 :- B0),R,CR).
'$assertat_d'(last,Head,Body,C0,R) :-
		'$compile_dynamic'((Head:-Body),0,CR),
		'$head_and_body'(C0, H0, B0),
		'$recordzp'(Head,(H0 :- B0),R,CR).

'$assert1'(last,C,H) :- '$$compile_stat'(C,C,0,H).
'$assert1'(first,C,H) :- '$$compile_stat'(C,C,2,H).

'$assertz_dynamic'(X,C,C0) :- (X/\4)=:=0, !,
	'$head_and_body'(C,H,B),
	'$assertat_d'(last,H,B,C0,_).
'$assertz_dynamic'(_,C,C0) :- 
	'$head_and_body'(C,H,B), functor(H,N,A),
	( '$check_if_reconsulted'(N,A) ->
		true
		 ;
		'$inform_as_reconsulted'(N,A),
		'$remove_all_d_clauses'(H)
	),
	'$assertat_d'(last,H,B,C0,_).

'$remove_all_d_clauses'(H) :-
	'$recordedp'(H,_,R), erase(R), fail.
'$remove_all_d_clauses'(_).


asserta(V,R) :- var(V), !,
	throw(error(instantiation_error,asserta(V,R))).
asserta(M:V,R) :- !, call(M:asserta(V,R)).
asserta(C,R) :- '$check_head_and_body'(C,H,B,asserta(C,R)),
	'$is_dynamic'(H), !,
	'$assertat_d'(first,H,B,C,R).
asserta(C,R) :- '$head_and_body'(C,H,B),
	'$undefined'(H),
	'$get_value'('$full_iso',true), !,
	functor(H, Na, Ar),
	dynamic(Na/Ar),
	'$assertat_d'(first,H,B,C,R).
asserta(C,R) :-
	'$head_and_body'(C,H,_),
	functor(H,Na,Ar),
	throw(error(permission_error(modify,static_procedure,Na/Ar),asserta(C,R))).

assertz(V,R) :- var(V), !,
	throw(error(instantiation_error,assertz(V,R))).
assertz(M:V,R) :- !, call(M:assertz(V,R)).
assertz(C,R) :- '$check_head_and_body'(C,H,B,assertz(C,R)),
	'$is_dynamic'(H), !,
	'$assertat_d'(last,H,B,C,R).
assertz(C,R) :- '$head_and_body'(C,H,B),
	'$undefined'(H), 
	'$get_value'('$full_iso',true), !,
	functor(H, Na, Ar),
	dynamic(Na/Ar),
	'$assertat_d'(last,H,B,C,R).
assertz(C,R) :-
	'$head_and_body'(C,H,_),
	functor(H,Na,Ar),
	throw(error(permission_error(modify,static_procedure,Na/Ar),asserta(C,R))).

assert(V,R) :- var(V), !,
	throw(error(instantiation_error,assert(V,R))).
assert(M:V,R) :- !, call(M:assert(V,R)).
assert(C,R) :- '$check_head_and_body'(C,H,B,assert(C,R)),
	'$is_dynamic'(H),
	'$get_value'('$full_iso',true), !,
	'$assertat_d'(last,H,B,C,R).
assert(C,R) :- '$head_and_body'(C,H,B),
	'$undefined'(H), !,
	functor(H, Na, Ar),
	dynamic(Na/Ar),
	'$assertat_d'(last,H,B,C,R).
assert(C,R) :-
	'$head_and_body'(C,H,_),
	functor(H,Na,Ar),
	throw(error(permission_error(modify,static_procedure,Na/Ar),assert(C,R))).

clause(V,Q) :- var(V), !, 
	throw(error(instantiation_error,clause(V,Q))).
clause(C,Q) :- number(C), !,
	throw(error(type_error(callable,C),clause(C,Q))).
clause(R,Q) :- db_reference(R), !,
	throw(error(type_error(callable,R),clause(R,Q))).
clause(M:P,Q) :- !, call(M:clause(P,Q)).
clause(P,Q) :- '$is_dynamic'(P), !,
	 '$recordedp'(P,(P:-Q),_). 
clause(P,Q) :-	
	'$some_recordedp'(P), !,
	 '$recordedp'(P,(P:-Q),_).
clause(P,Q) :-
	( '$system_predicate'(P) -> true ;
	    '$number_of_clauses'(P,N), N > 0 ),
	functor(P,Name,Arity),
	throw(error(permission_error(access,private_procedure,Name/Arity),
	      clause(P,Q))).

clause(M:P,Q,R) :- !, call(M:clause(P,Q,R)).
clause(P,Q,R) :-
	 ( '$is_dynamic'(P) ->
	 	'$recordedp'(P,(P:-Q),R)
	 ;
		throw(error(permission_error(access,private_procedure,N/A),
			clause(P,Q,R)))
	 ).

retract(V) :- var(V), !,
	throw(error(instantiation_error,retract(V))).
retract(M:C) :- !, call(M:retract(C)).
retract(C) :- '$check_head_and_body'(C,H,B,retract(C)),
	'$is_dynamic'(H), !,
	'$recordedp'(H,(H:-B),R), erase(R).
retract(C) :-
	'$fetch_predicate_indicator_from_clause'(C, PI),
	throw(error(permission_error(modify,static_procedure,PI),retract(C))).

retract(V,R) :- var(V), !,
	throw(error(instantiation_error,retract(V,R))).
retract(M:C,R) :- !,
	call(M:retract(C,R)).
retract(C,R) :-
	'$check_head_and_body'(C,H,B,retract(C,R)),
	db_reference(R), '$is_dynamic'(H), !,
	instance(R,(H:-B)), erase(R).
retract(C,R) :-
	'$head_and_body'(C,H,B,retract(C,R)),
	'$is_dynamic'(H), !,
	var(R),
	'$recordedp'(H,(H:-B),R),
	erase(R).
retract(C,R) :-
	'$fetch_predicate_indicator_from_clause'(C, PI).
	throw(error(permission_error(modify,static_procedure,PI),retract(C))).

'$fetch_predicate_indicator_from_clause'((C :- _), Na/Ar) :- !,
	functor(C, Na, Ar).
'$fetch_predicate_indicator_from_clause'(C, Na/Ar) :-
	functor(C, Na, Ar).
	

retractall(V) :- var(V), !,
	throw(error(instantiation_error,retract(V))).
retractall(M:V) :- !, call(M:retractall(V)).
retractall(T) :- '$undefined'(T),
	functor(T, Na, Ar),
	dynamic(Na/Ar),
	fail.
retractall(T) :- \+ '$is_dynamic'(T), !,
	functor(T,Na,Ar),
	throw(error(permission_error(modify,static_procedure,Na/Ar),retractall(T))).
retractall(T) :-
	'$erase_all_clauses_for_dynamic'(T).

'$erase_all_clauses_for_dynamic'(T) :-
	'$recordedp'(T,(T :- _),R), erase(R), fail.
'$erase_all_clauses_for_dynamic'(_).

abolish(N,A) :- var(N), !,
	throw(error(instantiation_error,abolish(N,A))).
abolish(N,A) :- var(A), !,
	throw(error(instantiation_error,abolish(N,A))).
abolish(M:N,A) :- !, call(M:abolish(N,A)).
abolish(N,A) :- 
	( '$recorded'('$predicate_defs','$predicate_defs'(N,A,_),R) -> erase(R) ),
	fail.
abolish(N,A) :- functor(T,N,A),
		( '$is_dynamic'(T) -> '$abolishd'(T) ;
	      	 /* else */	      '$abolishs'(T) ).

abolish(X) :- 
	get_value('$full_iso', true), !,
	'$new_abolish'(X).
abolish(X) :- 
	'$old_abolish'(X).

'$new_abolish'(V) :-
	'$check_error_in_predicate_indicator'(V, abolish(V)), !.
'$new_abolish'(Mod:PredSpec) :-
	call(Mod:'$new_abolish'(PredSpec)).
'$new_abolish'(Na/Ar) :-
	functor(Na, Ar, H),
	'$is_dynamic'(H), !,
	'$abolishd'(H).
'$new_abolish'(Pred) :-
	functor(H,Na,Ar),
	throw(error(permission_error(modify,static_procedure,Na/Ar),abolish(Pred))).


'$check_error_in_predicate_indicator'(V, Msg) :-
	var(V), !,
	throw(error(instantiation_error, Msg)).
'$check_error_in_predicate_indicator'(M:S, Msg) :- !,
	'$check_error_in_module'(M, Msg),
	'$check_error_in_predicate_indicator'(S, Msg).
'$check_error_in_predicate_indicator'(S, Msg) :-
	S \= _/_, !,
	throw(error(type_error(predicate_indicator,S), Msg)).
'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
	var(Na), !,
	throw(error(instantiation_error, Msg)).
'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
	\+ atom(Na), !,
	throw(error(type_error(atom,Na), Msg)).
'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
	var(Ar), !,
	throw(error(instantiation_error, Msg)).
'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
	\+ integer(Ar), !,
	throw(error(type_error(integer,Ar), Msg)).
'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
	Ar < 0, !,
	throw(error(domain_error(not_less_than_zero,Ar), Msg)).
% not yet implemented!
%'$check_error_in_predicate_indicator'(Na/Ar, Msg) :-
%	Ar < maxarity, !,
%	throw(error(type_error(representation_error(max_arity),Ar), Msg)).

'$check_error_in_module'(M, Msg) :-
	var(M), !,
	throw(error(instantiation_error, Msg)).
'$check_error_in_module'(M, Msg) :-
	\+ atom(M), !,
	throw(error(type_error(atom,Na), Msg)).

'$old_abolish'(V) :-
	'$check_error_in_predicate_indicator'(V, abolish(V)).
'$old_abolish'(M:N) :- !, call(M:abolish(N)).
'$old_abolish'([]) :- !.
'$old_abolish'([H|T]) :- !,  abolish(H), abolish(T).
'$old_abolish'(N/A) :- abolish(N,A).
	

'$abolishd'(T) :- '$recordedp'(T,_,R), erase(R), fail.
'$abolishd'(T) :- '$kill_dynamic'(T), fail.
'$abolishd'(_).

'$abolishs'(G) :- '$in_use'(G), !,
	functor(G,Name,Arity),
	throw(error(permission_error(modify,static_procedure_in_use,Name/Arity),abolish(G))).
'$abolishs'(G) :- '$system_predicate'(G), !,
	functor(G,Name,Arity),
	throw(error(permission_error(modify,static_procedure,Name/Arity),abolish(G))).
'$abolishs'(G) :-
	'$purge_clauses'(G),
	'$recordedp'(G,_,R), erase(R), fail.
'$abolishs'(_).


dynamic(X) :- var(X), !,
	throw(error(instantiation_error,dynamic(X))).
dynamic((A,B)) :- !, dynamic(A), dynamic(B).
dynamic(A/N) :- integer(N), atom(A), !, functor(T,A,N), $flags(T,F,F),
	X is F/\8'157000,
	(  X=:=0 -> $flags(T,F,8'20000);
	    '$is_dynamic'(T)  -> true;
	    F =:= 8'2200 , '$undefined'(T) -> '$flags'(T,F,8'20000);
	    throw(error(permission_error(modify,static_procedure,A/N),dynamic(A/N)))
	).
dynamic([]) :- !.
dynamic([H|L]) :- !, dynamic(H), dynamic(L).
dynamic(X) :- 
	throw(error(type_error(callable,X),dynamic(X))).
