/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		save.c							 *
* Last rev:								 *
* mods:									 *
* comments:	saving and restoring a Prolog computation		 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "@(#)save.c	1.3 3/15/90";
#endif

#include "Yap.h"
#include "alloc.h"
#include "clause.h"
#include "io.h"

#include <fcntl.h>
#include <unistd.h>
#ifndef FILENAME_MAX
#define FILENAME_MAX 1024 /* This is ok for Linux, should be ok for everyone */
#endif


/*********  hack for accesing several kinds of terms. Should be cleaned **/

#define AbsTerm(V) ((Term) (V))

     typedef union PTRS {
       CELL            asUnsigned;
       CELL           *asCellP;
       BITS16         *asWordP;
       DISPREG        *asDisplP;
       OPCODE         *asOpCodeP;
       OPREG          *asOpRegP;
       CODEADDR        asCodeP;
       SMALLUNSGN     *asSmallP;
       CODEADDR       *asPCodeP;
       Functor        *asFunctorP;
       Atom           *asAtomP;
       Term           *asTermP;
     }

     ptrs;

#define AsUnsigned(V) ((V).asUnsigned)
#define AsCellPtr(V) ((V).asCellP)
#define AsWordPtr(V) ((V).asWordP)
#define AsDisplPtr(V) ((V).asDisplP)
#define AsOpCodePtr(V) ((V).asOpCodeP)
#define AsOpRegPtr(V) ((V).asOpRegP)
#define AsCodePtr(V) ((V).asCodeP)
#define AsPtrCodePtr(V) ((V).asPCodeP)
#define AsSmallPtr(V) ((V).asSmallP)
#define AsFunctorPtr(V) ((V).asFunctorP)
#define AsAtomPtr(V)	((V).asAtomP)
#define AsTermPtr(V)	((V).asTermP)

#define DO_EVERYTHING 0
#define DO_ONLY_CODE  1

extern char     StartUpFile[];

static char end_msg[256] ="*** End of YAP saved state *****";

static int hp_size, gb_size, lc_size, tr_size;

#ifdef DEBUG

/*
 * FOR DEBUGGING define DEBUG_RESTORE0 to check the file stuff,
 * define DEBUG_RESTORE1 to see if it is able to prepare the chain,
 * define DEBUG_RESTORE2 to see how things are going,
 * define DEBUG_RESTORE3 to check if the atom chain is still a working
 * chain,
 * define DEBUG_RESTORE4 if you want to set the output for some
 * particular file,
 * define DEBUG_RESTORE5 if you want to see how the stacks are being
 * cleaned up,
 * define DEBUG_RESTORE6 if you want to follow the execution in 
 *
 * Also a file is defined where you can write things, by default stderr 
 *
 * Good Luck 
 */

#endif

STATIC_PROTO(void  myread, (int, char *, Int));
STATIC_PROTO(void  mywrite, (int, char *, Int));
STATIC_PROTO(int   open_file, (char *, int));
STATIC_PROTO(void  close_file, (void));
STATIC_PROTO(void  putout, (CELL));
STATIC_PROTO(void  putcellptr, (CELL *));
STATIC_PROTO(CELL  get_cell, (void));
STATIC_PROTO(CELL  *get_cellptr, ( /* CELL * */ void));
STATIC_PROTO(void  put_info, (int));
STATIC_PROTO(Int   p_save, (void));
STATIC_PROTO(int   check_header, (int));
STATIC_PROTO(void  get_heap_info, (void));
STATIC_PROTO(void  get_heap_regs, (void));
STATIC_PROTO(void  get_regs, (int));
STATIC_PROTO(void  get_insts, (void));
STATIC_PROTO(void  get_hash, (void));
STATIC_PROTO(void  CopyCode, (void));
STATIC_PROTO(void  CopyStacks, (void));
STATIC_PROTO(void  SkipStacks, (void));
STATIC_PROTO(void  get_coded, (int));
STATIC_PROTO(void  restore_codes, (void));
STATIC_PROTO(void  ConvDBList, (Term));
STATIC_PROTO(Term  AdjustDBTerm, (Term));
STATIC_PROTO(void  RestoreDB, (DBEntry *));
STATIC_PROTO(int  PrepareCCode, (void));
STATIC_PROTO(CELL  CCodeAdjust, (CELL));
STATIC_PROTO(void  RestoreClause, (CODEADDR));
STATIC_PROTO(void  CleanClauses, (CODEADDR, CODEADDR));
STATIC_PROTO(void  rehash, (CELL *, int, int));
STATIC_PROTO(void  RestoreIndex, (CODEADDR));
STATIC_PROTO(void  CleanCode, (PredEntry *));
STATIC_PROTO(void  RestoreEntries, (PropEntry *));
STATIC_PROTO(void  RestoreInvisibleAtoms, (void));
STATIC_PROTO(void  RestoreFreeSpace, (void));
STATIC_PROTO(void  restore_heap, (void));
STATIC_PROTO(void  RestoreState, (void));
#ifdef DEBUG_RESTORE3
STATIC_PROTO(void  ShowAtoms, (void));
STATIC_PROTO(void  ShowEntries, (PropEntry *));
#endif
STATIC_PROTO(int   OpenRestore, (char *, int));
STATIC_PROTO(void  CloseRestore, (void));
STATIC_PROTO(void  RestoreHeap, (void));
STATIC_PROTO(Int  p_restore, (void));
STATIC_PROTO(void  restore_heap_regs, (void));
STATIC_PROTO(void  restore_regs, (int));
STATIC_PROTO(void  PrepareEntries, (PropEntry *));
STATIC_PROTO(void  PrepareAtoms, (void));
STATIC_PROTO(void  ConvDBStruct, (Term));
STATIC_PROTO(Term  AdjustAppl, (Term));
STATIC_PROTO(Term  AdjustPair, (Term));
#ifdef MACYAP
STATIC_PROTO(void NewFileInfo, (long, long));
extern int      DefVol;
#endif


#ifdef LIGHT

#include <unix.h>
#include <strings.h>

void 
LightBug(char *);

static void 
LightBug(s)
	char           *s;
{
}

#endif				/* LIGHT */

#if SHORT_INTS

#ifdef M_WILLIAMS
#include <fcntl.h>
#endif

static void 
myread(fd, buff, len)
	int             fd;
	char           *buff;
	Int             len;
{
  while (len > 16000) {
    int nchars = read(fd, buff, 16000);
    if (nchars <= 0)
      Error(FATAL_ERROR,TermNil,"bad saved state, system corrupted");
    len -= 16000;
    buff += 16000;
  }
  read(fd, buff, (unsigned) len);
}

static void 
mywrite(fd, buff, len)
	int             fd;
	char           *buff;
	long            len;
{
	while (len > 16000) {
		write(fd, buff, 16000);
		len -= 16000;
		buff += 16000;
	}
	write(fd, buff, (unsigned) len);
}

#else				/* SHORT_INTS */

#if SUN
#include <sys/file.h>
#endif

inline static
void myread(int fd, char *buffer, Int len) {
  int nread;

  while (len > 0) {
    nread = read(fd, buffer, (int)len);
    if (nread < 1) {
      Error(FATAL_ERROR,TermNil,"bad saved state, system corrupted");
    }
    buffer += nread;
    len -= nread;
  }
}

inline static
void mywrite(int fd, char *buff, Int len) {
  Int nwritten;

  while (len > 0) {
    nwritten = (Int)write(fd, buff, (int)len);
    if (nwritten == -1) {
      Error(SYSTEM_ERROR,TermNil,"write error while saving");
    }
    buff += nwritten;
    len -= nwritten;
  }
}

#endif				/* SHORT_INTS */

#define FullSaved		1

#define CharP(X)	((char *) (X))
/* When wanting to adjust we use */
#define Adjust(X)	((char *)((X) ? Unsigned(X)+HDiff : NIL ))
#if	USE_OFFSETS
#define AtomAdjust(X)	((Atom)(X))
#define FuncAdjust(X)	((Functor)Adjust(X))
#define PropAdjust(X)	(X)
#else
#define AtomAdjust(X)	((Atom)(Unsigned(X) + HDiff))
#define FuncAdjust(X)	((Functor)(Unsigned(X) + FuncDiff))
#define PropAdjust(X)	((Prop)((X) ? Unsigned(X)+HDiff : NIL))
#endif
#define DBRefAdjust(X)	((DBRef)((X) ? Unsigned(X)+HDiff : NIL))
#define CodeAdjust(X)	((CODEADDR)(Unsigned(X) + HDiff))
#define BlockAdjust(X)	((BlockHeader *)((X) ? Unsigned(X) + HDiff : NIL))
#define PToGloAdjust(X) ((CELL *)(Unsigned(X)+GDiff))
#define PToTRAdjust(X)	((CELL *)(Unsigned(X)+TRDiff))
#define PToLocAdjust(X)	((CELL *)(Unsigned(X)+LDiff))
#define PToHAdjust(X)	((CELL *)(Unsigned(X)+HDiff))
#define GlobalAdjust(X) (Unsigned(X)+GDiff)
#define LocalAdjust(X)	(Unsigned(X)+LDiff)
#define HeapAdjust(X)	(Unsigned(X)+HDiff)
#define RegAdjust(X)	(Unsigned(X) + XDiff)
#define TrailAdjust(X)	(Unsigned(X)+TRDiff)


#define IsOldLocal(X)	( CharP(X) <= CharP(OldLCL0) && CharP(X) >= CharP(OldASP))
#define IsOldGlobal(X)	( CharP(X) >= CharP(OldH0) && CharP(X) <= CharP(OldH) )
#define IsOldTrail(X)	( CharP(X) <= CharP(OldTR) && CharP(X) >= CharP(OldTrailBase) )
#define IsOldCode(X)	( CharP(X)<CharP(OldHeapTop)&&CharP(X)>=CharP(OldHeapBase))
#define IsOldHeap(X)	IsOldCode(X)


#define Atomics		0
#define Funcs		1

/* Where the code was before */



typedef CELL   *CELLPOINTER;

int             splfild = 0;

#ifdef DEBUG

#ifdef DEBUG_RESTORE4
static FILE    *errout;
#else
#define errout YP_stderr
#endif

#endif				/* DEBUG */


static CELL     HDiff, LDiff, GDiff, XDiff, TRDiff;

#if USE_OFFSETS == 0
static CELL     FuncDiff;
#endif

static CELL    *OldASP, *OldLCL0, *OldXREGS;
static tr_fr_ptr OldTR;
static CELL    *OldH0, *OldH;

static CELL    *HCCode, *HCStart, *HCEnd;

static ADDR     OldHeapBase, OldHeapTop;

static ADDR     OldTrailBase;

static struct various_codes *old_heap_regs;

static Int      OldHeapUsed;

static ptrs     Base;

static BlockHeader *OldFreeBlocks;

/* Recovering old instructions */

#if USE_THREADED_CODE
/* Points to the start of the hash table where the old opcode numbers are */
static OPCODE   *ops;

#define oldop(X)	(ops[(0x3fffL & ((X)>>2))])
#else
#define oldop(X)	(X)
#endif

static Atom     Aenhanced;

/* Open a file to read or to write */
static int 
open_file(char *ss, int flag)
{
  int splfild;

#ifdef M_WILLIAMS
  if (flag & O_CREAT)
    splfild = creat(ss, flag);
  else
    splfild = open(ss, flag);
  if (splfild < 0) {
#else
#ifdef O_BINARY
    if ((splfild = open(ss, flag | O_BINARY), 0755) < 0)
#else
    if ((splfild = open(ss, flag, 0755)) < 0) 
#endif
#endif				/* M_WILLIAMS */
      {
	splfild = 0;	/* We do not have an open file */
	if (PrologMode == BootMode) {
	  return(-1);
	} else {
	  if (flag == O_RDONLY) {
	    Error(SYSTEM_ERROR,MkAtomTerm(LookupAtom(ss)),
		  "restore/1, open(%s)", strerror(errno));
	    return(-1);
	  } else {
	    Error(SYSTEM_ERROR,MkAtomTerm(LookupAtom(ss)),
		  "save/1, open(%s)", strerror(errno));
	    return(-1);
	  }
	}
      }
#ifdef undf0
      YP_fprintf(errout, "Opened file %s\n", ss);
#endif
      return(splfild);
}

static void 
close_file(void)
{
  close(splfild);
  splfild = 0;
}

/* stores a cell in a file */
static void 
putout(CELL l)
{
	mywrite(splfild, (char *) &l, sizeof(CELL));
}

/* stores a pointer to a cell in a file */
static void 
putcellptr(CELL *l)
{
	mywrite(splfild, (char *) &l, sizeof(CELLPOINTER));
}

/* gets a cell from a file */
static CELL 
get_cell(void)
{
	CELL            l;
	myread(splfild, (char *) &l, Unsigned(sizeof(CELL)));
	return (l);
}

/* gets a pointer to cell from a file */
static CELL    *
get_cellptr(void)
{
	CELL           *l;
	myread(splfild, (char *) &l, Unsigned(sizeof(CELLPOINTER)));
	return (l);
}

/*
 * writes the header (at the moment YAPV*), info about what kind of saved
 * set, the work size, and the space ocuppied 
 */
static void 
put_info(int info)
{
 char     msg[256];

 sprintf(msg, "#!/bin/sh\nexec %s/yap.exe $0 -- $*\n%cYAPV%s", BIN_DIR, 1, version_number);
 mywrite(splfild, msg, strlen(msg) + 1);
 putout(Unsigned(info));
 /* current state of stacks, to be used by SavedInfo */
#if defined(YAPOR) || defined(TABLING)
 /* space available in heap area */
 putout(Unsigned(GlobalBase)-Unsigned(HeapBase)-global_data_size()-local_data_size());
 /* space available for stacks */
 putout(Unsigned(LocalBase)-Unsigned(GlobalBase)+CellSize);
#else
 /* space available in heap area */
 putout(Unsigned(GlobalBase)-Unsigned(HeapBase));
 /* space available for stacks */
 putout(Unsigned(LocalBase)-Unsigned(GlobalBase));
#endif /* YAPOR || TABLING */
 /* space available for trail */
 putout(Unsigned(TrailTop)-Unsigned(TrailBase));
 /* Space used in heap area */
 putout(Unsigned(HeapTop)-Unsigned(HeapBase));
 /* Space used for local stack */
 putout(Unsigned(LCL0)-Unsigned(ASP));
 /* Space used for global stack */
 putout(Unsigned(H) - Unsigned(H0));
 /* Space used for trail */
 putout(Unsigned(TR) - Unsigned(TrailBase));
}

/* Saves a complete prolog environment */
static Int 
p_save(void)
{
  char            s[FILENAME_MAX];
  register long   i, j;
#ifdef MACYAP
  NewFileInfo('YAPS', 'MYap');
#endif
  Term t1 = Deref(ARG1);
  if (!GetName(s, t1)) {
    Error(TYPE_ERROR_LIST,t1,"save/1");
    return(FALSE);
  }
  CloseStreams(TRUE);
  if ((splfild = open_file(s, O_WRONLY | O_CREAT)) < 0)
    return(FALSE);
  put_info(FullSaved);
  /* save all registers */
#ifdef YAP_ARRAYS
  putout((CELL)compile_arrays);
#endif
#ifdef OLD_REGS_SCHEME
  mywrite(splfild, (char *)&REGS, REG_SIZE *
	  sizeof(CELLPOINTER));
#else
  putcellptr((CELL *)CP);
  putcellptr(ENV);
  putcellptr(ASP);
  /* putout((CELL)N); */
  putcellptr(H0);
  putcellptr(LCL0);
  putcellptr(H);
  putcellptr(HB);
  putcellptr((CELL *)B);
  putcellptr((CELL *)TR);
  putcellptr(YENV);
  putcellptr(S);
  putcellptr((CELL *)P);
  putcellptr((CELL *)HeapPlus);
  putcellptr((CELL *)MyTR);
  putout(CreepFlag);
  putcellptr(CreepCode);
  putcellptr(UndefCode);
  putcellptr(SpyCode);
  putout(FlipFlop);
  putout(TermDot);
  putout(TermNil);
#ifdef COROUTINING
  putout(WokenGoals);
#ifdef MULTI_ASSIGNMENT_VARIABLES
  putcellptr(WakeUpCode);
  putout(MutableList);
#endif
#endif
#ifdef  DEPTH_LIMIT
  putout(DEPTH);
#endif
  mywrite(splfild, (char *)&XREGS[0], sizeof(CELL)*MaxTemps);
#endif
  /* The operand base */
  putcellptr(CellPtr(XREGS));
  /* And the fail adress */
  putcellptr(CellPtr(heap_regs));
  /* Now start by saving the code */
  /* the heap boundaries */
  putcellptr(CellPtr(HeapBase));
  putcellptr(CellPtr(HeapTop));
  /* and the space it ocuppies */
  putout(Unsigned(HeapUsed));
  /* Then the start of the free code */
  putcellptr(CellPtr(FreeBlocks));
  /* put the old trail base, just in case it moves again */
  putcellptr(CellPtr(TrailBase));
  /* Then the instructions */
  {
    OPCODE my_ops[_std_top+1];
    for (i = 0; i <= _std_top; ++i)
      my_ops[i] = opcode((int) i);
    mywrite(splfild, (char *)my_ops, sizeof(OPCODE)*(_std_top+1));
  }
  /* Then the module information */
  putout(Unsigned(NoOfModules));
  mywrite(splfild, (char *) ModuleName, NoOfModules * CellSize);
  /* The hash table */
  mywrite(splfild, (char *) HashChain, MaxHash * CellSize);
  /* and the current character codes */
  mywrite(splfild, chtype, NUMBER_OF_CHARS);
  /* Then save the whole heap */
  j = Unsigned(HeapTop) - Unsigned(HeapBase);
  mywrite(splfild, (char *) HeapBase, j);
  /* Now, go and save the state */
  /* Save the local stack */
  j = Unsigned(LCL0) - Unsigned(ASP);
  mywrite(splfild, (char *) ASP, j);
  /* Save the global stack */
  j = Unsigned(H) - Unsigned(H0);
  mywrite(splfild, (char *) H0, j);
  /* Save the trail */
  j = Unsigned(TR) - Unsigned(TrailBase);
  mywrite(splfild, (char *) TrailBase, j);
  /* Save a CRC */
  mywrite(splfild, end_msg, 256);
#ifdef MACYAP
  NewFileInfo('TEXT', 'MYap');
  if (DefVol)
    SetVol(0l, DefVol);
  DefVol = 0;
#endif
  close_file();
  return (TRUE);
}

/* Now, to restore the saved code */

/* First check out if we are dealing with a valid file */
static int 
check_header(int flag)
{
  char            pp[80];
  char msg[256];
  
  /* skip the first line */
  do {
    myread(splfild, pp, 1);
  } while (pp[0] != 1);
  /* now check the version */
  sprintf(msg, "YAPV%s", version_number);
  myread(splfild, pp, Unsigned(strlen(msg) + 1));
  if (strcmp(pp, msg) != 0) {
    Error(SYSTEM_ERROR,TermNil,"not a saved Prolog state");
    return(FALSE);
  }
  /* check info on header */
  /* ignore info on saved state */
  get_cell();
  /* ignore info on stacks size */
  get_cell();
  get_cell();
  get_cell();
  /* now, check whether we got enough enough space to load the
     saved space */
  if ((hp_size = get_cell()) > Unsigned(AuxTop) - Unsigned(HeapBase)) {
    Error(SYSTEM_ERROR,TermNil,"out of heap space");
    return(FALSE);
  }
  if ((lc_size = get_cell())+(gb_size=get_cell()) > Unsigned(LocalBase) - Unsigned(GlobalBase)) {
    Error(SYSTEM_ERROR,TermNil,"out of stack space");
    return(FALSE);
  }
  if ((tr_size = get_cell()) > Unsigned(TrailTop) - Unsigned(TrailBase)) {
    Error(SYSTEM_ERROR,TermNil,"out of trail space");
    return(FALSE);
  }
  return(TRUE);
}

/* Gets the state of the heap, and evaluates the related variables */
static void 
get_heap_info(void)
{
	old_heap_regs = (struct various_codes *)get_cellptr();
	OldHeapBase = (ADDR) get_cellptr();
	OldHeapTop = (ADDR) get_cellptr();
	OldHeapUsed = (Int) get_cell();
	OldFreeBlocks = (BlockHeader *) get_cellptr();
	HDiff = Unsigned(HeapBase) - Unsigned(OldHeapBase);
#if USE_OFFSETS==0
	FuncDiff = HDiff;
#endif
}

/* Gets some registers which do depend on the heap */
static void 
get_heap_regs(void)
{
	HeapPlus = (ADDR) get_cellptr();
	UndefCode = get_cellptr();
	CreepCode = get_cellptr();
	SpyCode = get_cellptr();
}

/* Gets the register array */
/* Saves the old bases for the work areas */
/* and evaluates the difference from the old areas to the new ones */
static void 
get_regs(int flag)
{
  CELL           *NewH0 = H0;
  CELL           *NewLCL0 = LCL0;
  /* Get regs */
#ifdef YAP_ARRAYS
  compile_arrays = (int)get_cell();
#endif
  if (flag == DO_EVERYTHING) {
    CP = (yamop *)get_cellptr();
    ENV = get_cellptr();
    ASP = get_cellptr();
    /* N = get_cell(); */
    H0 = get_cellptr();
    LCL0 = get_cellptr();
    H = get_cellptr();
    HB = get_cellptr();
    B = (choiceptr)get_cellptr();
    TR = (tr_fr_ptr)get_cellptr();
    YENV = get_cellptr();
    S = get_cellptr();
    P = (yamop *)get_cellptr();
    HeapPlus = (ADDR)get_cellptr();
    MyTR = (tr_fr_ptr)get_cellptr();
    CreepFlag = get_cell();
    CreepCode = get_cellptr();
    UndefCode = get_cellptr();
    SpyCode = get_cellptr();
    FlipFlop = get_cell();
    TermDot = get_cell();
    TermNil = get_cell();
#ifdef COROUTINING
    WokenGoals = get_cell();
#ifdef MULTI_ASSIGNMENT_VARIABLES
    WakeUpCode = get_cellptr();
    MutableList = get_cell();
#endif
#endif
#ifdef  DEPTH_LIMIT
    DEPTH = get_cell();
#endif
    myread(splfild, (char *)&XREGS[0], sizeof(CELL)*MaxTemps);
  /* Get the old bases */
    OldXREGS = get_cellptr();
  } else {
#ifdef COROUTINING
    lseek(splfild, sizeof(CELL), SEEK_CUR);
#ifdef MULTI_ASSIGNMENT_VARIABLES
    lseek(splfild, 2*sizeof(CELL), SEEK_CUR);
#endif
#endif
#ifdef  DEPTH_LIMIT
    lseek(splfild, sizeof(CELL), SEEK_CUR);
#endif
    lseek(splfild, (22+MaxTemps)*sizeof(CELL), SEEK_CUR);
  }
  get_heap_info();
  if (flag == DO_EVERYTHING) {
    /* get old trail base */
    OldTrailBase = (ADDR)get_cellptr();
    /* Save the old register where we can easily access them */
    OldASP = ASP;
    OldLCL0 = LCL0;
    OldH0 = H0;
    OldH = H;
    OldTR = TR;
    GDiff = Unsigned(NewH0) - Unsigned(H0);
    LDiff = Unsigned(NewLCL0) - Unsigned(LCL0);
    XDiff = Unsigned(XREGS) - Unsigned(OldXREGS);
    TRDiff = LDiff;
    H0 = NewH0;
    LCL0 = NewLCL0;
  }
}

/* Get the old opcodes and place them in a hash table */
static void 
get_insts(void)
{
  OPCODE old_ops[_std_top+1];
  
  myread(splfild, (char *)old_ops, sizeof(OPCODE)*(_std_top+1));
}

/* Get the old atoms hash table */
static void 
get_hash(void)
{
  NoOfModules = get_cell();
  myread(splfild, (char *) ModuleName, NoOfModules * CellSize);
  myread(splfild, (char *) HashChain, Unsigned(MaxHash * CellSize));
  myread(splfild, chtype , NUMBER_OF_CHARS);
}

/* Copy all of the old code to the new Heap */
static void 
CopyCode(void)
{
  myread(splfild, (char *) HeapBase,
	 Unsigned(OldHeapTop) - Unsigned(OldHeapBase));
}

/* Copy the local and global stack and also the trail to their new home */
/* In REGS we still have nonadjusted values !! */
static void 
CopyStacks(void)
{
  Int             j;
  char           *NewASP, *NewTR;

  j = Unsigned(OldLCL0) - Unsigned(ASP);
  NewASP = (char *) (Unsigned(ASP) + (Unsigned(LCL0) - Unsigned(OldLCL0)));
  myread(splfild, (char *) NewASP, j);
  j = Unsigned(H) - Unsigned(OldH0);
  myread(splfild, (char *) H0, j);
  j = Unsigned(TR) - Unsigned(OldTrailBase);
  NewTR = (char *) (Unsigned(TR) + Unsigned(TrailBase)
		    - Unsigned(OldTrailBase));
  myread(splfild, TrailBase, j);
}

static void 
SkipStacks(void)
{
  lseek(splfild, lc_size+gb_size+tr_size, SEEK_CUR);
}


/* get things which are saved in the file */
static void 
get_coded(int flag)
{
  char my_end_msg[256];
  get_regs(flag);
  get_insts();
  get_hash();
  CopyCode();
  if (flag == DO_EVERYTHING) {
    CopyStacks();
  } else {
    SkipStacks();
  }
  /* Check CRC */
  myread(splfild, my_end_msg, 256);
  if (strcmp(end_msg,my_end_msg) != 0)
    Error(FATAL_ERROR,TermNil,"bad saved state, system corrupted");
}

/* Now, everything on its place so you must adjust the pointers */

/* restore the failcodes */
static void 
restore_codes(void)
{
  heap_regs = (struct various_codes *)CodeAdjust(old_heap_regs);
#ifdef YAPOR
  heap_regs->getworkfirsttimecode.opc = opcode(_getwork_first_time);
  heap_regs->getworkcode.opc = opcode(_getwork);
  heap_regs->getworkcode.YAMOP_LTT = 0;
  heap_regs->getworkcode.YAMOP_CUT = 0;
  heap_regs->getworkcode.YAMOP_SEQ = FALSE;
  heap_regs->getworkcode_seq.opc = opcode(_getwork_seq);
  heap_regs->getworkcode_seq.YAMOP_LTT = 0;
  heap_regs->getworkcode_seq.YAMOP_CUT = 0;
  heap_regs->getworkcode_seq.YAMOP_SEQ = FALSE;
#endif /* YAPOR */
#ifdef TABLING
  heap_regs->tablecompletioncode.opc = opcode(_table_completion);
  heap_regs->tableanswerreturncode.opc = opcode(_table_answer_return);
#ifdef YAPOR
  heap_regs->tablecompletioncode.YAMOP_LTT = 0;
  heap_regs->tablecompletioncode.YAMOP_CUT = 0;
  heap_regs->tablecompletioncode.YAMOP_SEQ = FALSE;
  heap_regs->tableanswerreturncode.YAMOP_LTT = 0;
  heap_regs->tableanswerreturncode.YAMOP_CUT = 0;
  heap_regs->tableanswerreturncode.YAMOP_SEQ = FALSE;
#endif /* YAPOR */
#endif /* TABLING */
  heap_regs->failcode = opcode(_op_fail);
  heap_regs->trustfailcode = opcode(_trust_fail);
  heap_regs->yescode = opcode(_Ystop);
#ifdef YAPOR
  heap_regs->nocode.opc = opcode(_Nstop);
  heap_regs->nocode.YAMOP_LTT = 1;
  heap_regs->nocode.YAMOP_CUT = 0;
  heap_regs->nocode.YAMOP_SEQ = FALSE;
#else
  heap_regs->nocode = opcode(_Nstop);
#endif  /* YAPOR */

#ifdef YAPOR
  heap_regs->rtrycode.YAMOP_LTT = 1;
  heap_regs->rtrycode.YAMOP_CUT = 0;
  heap_regs->rtrycode.YAMOP_SEQ = FALSE;
#endif /* YAPOR */
  ((yamop *)(&heap_regs->rtrycode))->opc = opcode(_retry_and_mark);
  if (((yamop *)(&heap_regs->rtrycode))->u.ld.d != NIL)
    ((yamop *)(&heap_regs->rtrycode))->u.ld.d =
      CodeAdjust(((yamop *)(&heap_regs->rtrycode))->u.ld.d);
  {
    int             arity;
    arity = heap_regs->clausecode.arity;
    if (heap_regs->clausecode.clause != NIL)
      heap_regs->clausecode.clause =
	CodeAdjust(heap_regs->clausecode.clause);
    if (arity)
      heap_regs->clausecode.func = (Functor)
	FuncAdjust(heap_regs->clausecode.func); 
    else
      heap_regs->clausecode.func =
	(Functor)AtomAdjust(heap_regs->clausecode.func);
  }
  if (heap_regs->invisiblechain != NIL)
    heap_regs->invisiblechain =
      AtomAdjust(heap_regs->invisiblechain);
  
  /* restore consult stack. It consists of heap pointers, so it
     is easy to fix.
  */
  heap_regs->consultlow  = (consult_obj *) 
    AtomAdjust(heap_regs->consultlow);
  heap_regs->consultbase = (consult_obj *) 
    AtomAdjust(heap_regs->consultbase);
  heap_regs->consultsp   = (consult_obj *) 
    AtomAdjust(heap_regs->consultsp);
  {
    /* we assume all pointers have the same size */
    register consult_obj *pt = heap_regs->consultsp;
    while (pt <
	   heap_regs->consultlow+heap_regs->consultcapacity) {
      pt->p = (Prop)AtomAdjust(pt->p);
      pt ++;
    }
  }
}

/* restore some heap registers */
static void 
restore_heap_regs(void)
{
  HeapPlus = (ADDR) CodeAdjust(HeapPlus);
  HeapTop = (ADDR) CodeAdjust(OldHeapTop);
  *((SEG_SIZE *) HeapTop) = InUseFlag;
  HeapMax = HeapUsed = OldHeapUsed;
  restore_codes();
}

/* adjust abstract machine registers */
static void 
restore_regs(int flag)
{
  restore_heap_regs();
  if (flag == DO_EVERYTHING) {
    CP = (yamop *)PToHAdjust(CellPtr(CP));
    ENV = PToLocAdjust(ENV);
    ASP = PToLocAdjust(ASP);
    H = PToGloAdjust(H);
    B = (choiceptr)PToLocAdjust(CellPtr(B));
    TR = (tr_fr_ptr)PToTRAdjust(TR);
    P = (yamop *)PToHAdjust(CellPtr(P));
    HB = PToLocAdjust(HB);
    YENV = PToLocAdjust(YENV);
    S = PToGloAdjust(S);
    HeapPlus = (ADDR) CodeAdjust(HeapPlus);
    if (MyTR)
      MyTR = (tr_fr_ptr)PToTRAdjust(MyTR);
    if (UndefCode)
      UndefCode = PToHAdjust(UndefCode);
    if (CreepCode)
      CreepCode = PToHAdjust(CreepCode);
    if (SpyCode)
      SpyCode = PToHAdjust(SpyCode);
#ifdef COROUTINING
    if (WakeUpCode != NULL)
      WakeUpCode = PToHAdjust(WakeUpCode);
#ifdef MULTI_ASSIGNMENT_VARIABLES
    WokenGoals = AbsAppl(PToGloAdjust(RepAppl(WokenGoals)));
    MutableList = AbsAppl(PToGloAdjust(RepAppl(MutableList)));
#endif
#endif
  }
}

/*
 * Because of the C functions, one must  prepare the atoms and each of their
 * entries: this is done by restoring the pointers and by finding all C
 * predicates 
 */

static void 
PrepareEntries(PropEntry *pp)
{
	while (!EndOfPAEntr(pp)) {
		pp->NextOfPE = PropAdjust(pp->NextOfPE);
		if (!(pp->KindOfPE & 0x8000)
		    && (((PredEntry *) pp)->PredFlags & CPredFlag)) {
			*HCCode++ = (CELL) (((PredEntry *) pp)->CodeOfPred);
			*HCCode++ = (CELL) AbsProp(pp);
		}
		pp = RepProp(pp->NextOfPE);
	}
}

static void 
PrepareAtoms(void)
{
	Atom           *HashPtr = HashChain;
	register int    i;
	for (i = 0; i < MaxHash; ++i) {
		if (*(HashPtr)) {
			AtomEntry      *at;
			*HashPtr = AtomAdjust(*HashPtr);
			at = RepAtom(*HashPtr);
			do {
				if (at->NextOfAE)
					at->NextOfAE = AtomAdjust(at->NextOfAE);
				at->PropOfAE = PropAdjust(at->PropOfAE);
#ifdef DEBUG_RESTORE1
				YP_fprintf(YP_stderr, "%s\n", at->StrOfAE);
#endif
				PrepareEntries(RepProp(at->PropOfAE));
			} while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
		}
		HashPtr++;
	}
	/* And now prepare the protected chain */
	if (*INVISIBLECHAIN) {
		AtomEntry      *at = RepAtom(*INVISIBLECHAIN);
		do {
			if (at->NextOfAE)
				at->NextOfAE = AtomAdjust(at->NextOfAE);
			at->PropOfAE = PropAdjust(at->PropOfAE);
#ifdef DEBUG_RESTORE1
			YP_fprintf(YP_, "%s\n", at->StrOfAE);
#endif
			PrepareEntries(RepProp(at->PropOfAE));
		} while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
	}
}

/* Restoring the heap */

/* Converts a structure of the DB, as it was saved in the heap */
/* uses a variable base */
static void 
ConvDBStruct(Term Struct)
{
	ptrs            TermP;
	int             Arity, i;
	AsUnsigned(TermP) = AsUnsigned(Base) + (BITS32) RepAppl(Struct);
	*AsCellPtr(TermP) = (CELL) FuncAdjust(*AsCellPtr(TermP));
	Arity = ArityOfFunctor((Functor)(*AsCellPtr(TermP)++));
	for (i = 0; i < Arity; i++) {
		register Term   t = *AsTermPtr(TermP);
		if (IsAtomTerm(t))
			*AsAtomPtr(TermP) = AtomAdjust(t);
		else if (IsApplTerm(t))
			ConvDBStruct(t);
		else if (IsPairTerm(t))
			ConvDBList(t);
		AsCellPtr(TermP)++;
	}
}

/* Converts a list of the DB, as it was saved in the heap */
/* uses a variable base */
static void 
ConvDBList(Term List)
{
	ptrs            TermP;
	int             i;
	AsUnsigned(TermP) = AsUnsigned(Base) + (BITS32) RepPair(List);
	for (i = 0; i < 2; i++) {
		register Term   t = *AsTermPtr(TermP);
		if (IsAtomTerm(t))
			*AsAtomPtr(TermP) = AtomAdjust(t);
		else if (IsApplTerm(t))
			ConvDBStruct(t);
		else if (IsPairTerm(t))
			ConvDBList(t);
		AsCellPtr(TermP)++;
	}
}

/* adjusts terms stored in the data base, when they have no variables */
static Term 
AdjustDBTerm(Term trm)
{
	Term           *p;
	if (IsAtomTerm(trm))
		return ((Term) AtomAdjust(trm));
	if (IsPairTerm(trm)) {
		p = (CELL *)HeapAdjust((CELL)RepPair(trm));
		*p = AdjustDBTerm(*p);
		++p;
		*p = AdjustDBTerm(*p);
		return (AbsPair(p-1));
	}
	if (IsApplTerm(trm)) {
		int             Arity, i;
		Term *p0 = p = (CELL *)HeapAdjust((CELL)RepAppl(trm));
		*p = (Term) FuncAdjust(*p);
		Arity = ArityOfFunctor((Functor)(*p++));
		for (i = 0; i < Arity; ++i) {
			*p = AdjustDBTerm(*p);
			p++;
		}
		return (AbsAppl(p0));
	}
	if (IsRefTerm(trm))
		return ((Term) AtomAdjust(trm));
	return (trm);
}

/* Restores a DB structure, as it was saved in the heap */
static void 
RestoreDB(DBEntry *pp)
{
	register DBRef  dbr;
	pp->First = DBRefAdjust(pp->First);
	pp->Last = DBRefAdjust(pp->Last);
#ifndef KEEP_ENTRY_AGE
	pp->FirstNEr = DBRefAdjust(pp->FirstNEr);
#endif
	if (pp->ArityOfDB)
		pp->FunctorOfDB = FuncAdjust(pp->FunctorOfDB);
	else
		pp->FunctorOfDB = (Functor) AtomAdjust(pp->FunctorOfDB);
	dbr = pp->First;
	/* While we have something in the data base, restore it */
	while (dbr) {
#ifdef DEBUG_RESTORE
		YP_fprintf(errout, "Restoring at %x", dbr);
		if (dbr->Flags & DBAtomic)
			YP_fprintf(errout, " an atomic term\n");
		else if (dbr->Flags & DBNoVars)
			YP_fprintf(errout, " with no vars\n");
		else if (dbr->Flags & DBComplex)
			YP_fprintf(errout, " complex term\n");
		else if (dbr->Flags & DBIsRef)
			YP_fprintf(errout, " a ref\n");
		else
			YP_fprintf(errout, " a var\n");
#endif
		dbr->Parent = Adjust(dbr->Parent);
		if (dbr->Code != NIL)
			dbr->Code = CodeAdjust(dbr->Code);
		if (dbr->Flags & DBAtomic) {
			if (IsAtomTerm(dbr->Entry))
				dbr->Entry = (CELL) AtomAdjust(dbr->Entry);
		} else if (dbr->Flags & DBNoVars)
			dbr->Entry = (CELL) AdjustDBTerm((Term) dbr->Entry);
		else if (dbr->Flags & DBComplex) {
			AsTermPtr(Base) = dbr->Contents;
			AsCellPtr(Base)++;
			if (IsApplTerm((Term) dbr->Entry))
				ConvDBStruct((Term) dbr->Entry);
			else
				ConvDBList((Term) dbr->Entry);
		}
		dbr->Prev = DBRefAdjust(dbr->Prev);
		dbr->Next = DBRefAdjust(dbr->Next);
		if (dbr->Flags & DBWithRefs) {
			Term           *cp;
			Term            tm;
			cp = (Term *) ((CODEADDR) dbr + SizeOfBlock(CodePtr(dbr)));
			while ((tm = *--cp) != 0)
				*cp = (Term) HeapAdjust(tm);
		}
#ifdef DEBUG_RESTORE2
		YP_fprintf(errout, "Recomputing masks\n");
#endif
		if ((dbr->Flags & DBComplex) || (dbr->Flags & DBNoVars)) {
			dbr->Mask = EvalMasks((Term) dbr->Entry, &(dbr->Key));
		}
		dbr = dbr->Next;
	}
}

/* After initialazing, we have to get the new adresses for the C functions */
static int 
PrepareCCode(void)
{
	int             DirtyCode = FALSE;
	register CELL  *HCPtr = HCStart;
	while (HCPtr != HCCode) {
		HCPtr++;
		*HCPtr = (CELL) (RepPredProp((Prop)(*HCPtr))->CodeOfPred);
		if (*HCPtr != *(HCPtr - 1))
			DirtyCode = TRUE;
		HCPtr++;
	}
	*HCPtr++ = 0;		/* This is where sort should be */
	HCEnd = HCPtr;
	if (DirtyCode)
		return (FALSE);
	return (TRUE);
}

/* If we found a C function in the code, return its new adress */
static CELL 
CCodeAdjust(CELL CAdress)
{
	register CELL  *cptr = HCStart;
	while (*cptr != CAdress) {
		if (*cptr == 0)
			return (CAdress);
		cptr += 2;
	}
	return (*++cptr);
}

/* Restores a prolog clause, in its compiled form */
static void 
RestoreClause(CODEADDR Cl)
/*
 * Cl points to the start of the code, IsolFlag tells if we have a single
 * clause for this predicate or not 
 */
	                   
{
	ptrs            code, s_c;
	register int    op;
#ifdef DEBUG_RESTORE2
	CODEADDR        EndOfC;
#endif
	SMALLUNSGN      cl_type;
	AsCodePtr(s_c) = AsCodePtr(code) = Cl;
	AsCodePtr(s_c) -= SmallSize + CellSize;
#ifdef DEBUG_RESTORE2
	EndOfC = Cl + SizeOfBlock(Cl - SmallSize - CellSize) - SmallSize - CellSize;
#endif
	if (((cl_type = KindOfArg(AsCodePtr(code))) == ApplCl) ||
	    (cl_type == ListCl && KindOfListArg(AsCodePtr(code)) == ApplCl))
		*AsFunctorPtr(s_c) = FuncAdjust(*AsCellPtr(s_c));
	else if (((cl_type == AtCl) ||
		  (cl_type == ListCl && KindOfListArg(AsCodePtr(code)) == AtCl)) &&
		 IsAtomTerm(*AsCellPtr(s_c)))
		*AsAtomPtr(s_c) = AtomAdjust(*AsCellPtr(s_c));
	if ((*(AsSmallPtr(code) - 1)) & ErasedMask) {
#ifdef DEBUG_RESTORE2
		YP_fprintf(errout, "Restoring erased code\n");
#endif
		*AsOpCodePtr(code)++ = opcode(_jump);
		*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
		AsPtrCodePtr(code)++;
		*AsOpCodePtr(code)++ = opcode(_op_fail);
		*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
		return;
	}
#ifdef DEBUG_RESTORE2
	YP_fprintf(YP_stderr,"apanhou inicialmente %lx, 0x%lx, 0x%lx\n", *AsOpCodePtr(code),
	       Cl, EndOfC);
#endif
	/* Get the stored operator */
	do {
		op = oldop(*AsOpCodePtr(code));
#ifdef DEBUG_RESTORE2
		YP_fprintf(errout, "%d ", op);
#endif
		*AsOpCodePtr(code)++ = opcode((int) op);
		switch (op) {
		case _Ystop:
		case _Nstop:
#ifdef DEBUG_RESTORE2
			YP_fprintf(errout, "left OK\n");
#endif
			return;
		case _cut:
		case _cut_t:
		case _cut_e:
		case _or_last:
		case _pop:
		case _op_fail:
		case _trust_fail:
		case _write_list:
		case _unify_list:
		case _write_l_list:
		case _unify_l_list:
		case _procceed:
		case _undef_p:
		case _allocate:
		case _deallocate:
		case _unify_void:
		case _write_void:
			break;
		case _try_me:
		case _retry_me:
		case _try_me0:
		case _retry_me0:
		case _try_me1:
		case _retry_me1:
		case _try_me2:
		case _retry_me2:
		case _try_me3:
		case _retry_me3:
		case _try_me4:
		case _retry_me4:
		case _try_and_mark:
		case _retry_and_mark:
		case _or_else:
		case _spy_or_trymark:
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			if (op != _or_else)
				AsOpRegPtr(code)++;
			break;
		case _trust_me:
		case _trust_me0:
		case _trust_me1:
		case _trust_me2:
		case _trust_me3:
		case _trust_me4:
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			if (*AsCellPtr(code))
				*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			AsOpRegPtr(code)++;
			break;
		case _unify_n_voids:
		case _write_n_voids:
		case _get_list:
		case _put_list:
		case _unify_x_var:
		case _unify_y_var:
		case _unify_x_val:
		case _unify_y_val:
		case _unify_x_loc:
		case _unify_y_loc:
		case _write_x_var:
		case _write_y_var:
		case _write_x_val:
		case _write_y_val:
		case _write_x_loc:
		case _write_y_loc:
		case _save_b_x:
		case _save_b_y:
		case _comit_b_x:
		case _comit_b_y:
		case _save_appl_x:
		case _save_appl_y:
		case _save_pair_x:
		case _save_pair_y:
			AsOpRegPtr(code)++;
			break;
		case _unify_atom:
		case _write_atom:
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			if (!IsNumTerm(*AsCellPtr(code)))
				*AsAtomPtr(code) = AtomAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			break;
		case _unify_struct:
		case _write_struct:
		case _unify_l_struc:
		case _write_l_struc:
#ifdef SFUNC
		case _unify_s_f:
		case _write_s_f:
#endif
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			*AsFunctorPtr(code) = FuncAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			break;
#ifdef SFUNC
		case _unify_s_xvar:
		case _write_s_xvar:
		case _unify_s_xval:
		case _write_s_xval:
		case _unify_s_yvar:
		case _write_s_yvar:
		case _unify_s_yval:
		case _write_s_yval:
			AsCellPtr(code)++;
			AsCellPtr(code)++;
			break;
		case _unify_s_a:
	case_write_s_a:
			AsCellPtr(code)++;
			*AsCellPtr(code) = AtomAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			break;
		case _get_s_end:
		case _put_s_end:
		case _unify_s_end:
		case _write_s_end:
			AsCellPtr(code)++;
			AsCellPtr(code)++;
			break;
#endif
		case _jump:
		case _try_in:
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsCellPtr(code)++;
			break;
		case _execute:
		case _dexecute:
			/* if (EndOfC - AsCodePtr(code) < CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _pop_n:
			AsCellPtr(code)++;
			break;
		case _unify_n_atoms:
		case _write_n_atoms:
		case _get_atom:
		case _put_atom:
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			if (!IsNumTerm(*AsCellPtr(code)))
				*AsAtomPtr(code) = AtomAdjust(*AsAtomPtr(code));
			AsCellPtr(code)++;
			break;
		case _get_struct:
		case _put_struct:
#ifdef SFUNC
		case _get_s_f:
		case _put_s_f:
#endif
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			*AsFunctorPtr(code) = FuncAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			break;
		case _either:
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _call:
		case _fcall:
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _call_cpred:
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			*AsCellPtr(code) = CCodeAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			break;
		case _call_c_wfail:
			/*
			 * if (EndOfC - AsCodePtr(code) < CellSize+OpRegSize)
			 * return; 
			 */
			AsOpRegPtr(code)++;
			*AsCellPtr(code) = CCodeAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsCellPtr(code)++;
			break;
		case _unify_x_var2:
		case _get_x_var:
		case _get_y_var:
		case _get_x_val:
		case _get_y_val:
		case _put_x_var:
		case _put_y_var:
		case _put_x_val:
		case _put_y_val:
		case _glist_valx:
		case _glist_valy:
		case _gl_void_varx:
		case _gl_void_vary:
		case _gl_void_valx:
		case _gl_void_valy:
		case _put_unsafe:
			AsOpRegPtr(code)++;
			AsOpRegPtr(code)++;
			break;
		case _p_atom_x:
		case _p_atom_y:
		case _p_number_x:
		case _p_number_y:
		case _p_atomic_x:
		case _p_atomic_y:
		case _p_integer_x:
		case _p_integer_y:
		case _p_var_x:
		case _p_var_y:
		case _p_nonvar_x:
		case _p_nonvar_y:
		case _p_equal:
		case _p_dif:
		case _p_cut_by_x:
		case _p_cut_by_y:
		case _p_db_ref_x:
		case _p_db_ref_y:
		case _p_compound_x:
		case _p_compound_y:
		case _p_float_x:
		case _p_float_y:
		case _p_primitive_x:
		case _p_primitive_y:
		case _p_plus_vv:
		case _p_plus_vc:
		case _p_minus_vv:
		case _p_minus_cv:
		case _p_times_vv:
		case _p_times_vc:
		case _p_div_vv:
		case _p_div_cv:
		case _p_div_vc:
		case _p_and_vv:
		case _p_and_vc:
		case _p_or_vv:
		case _p_or_vc:
		case _p_sll_vv:
		case _p_sll_vc:
		case _p_sll_cv:
		case _p_slr_vv:
		case _p_slr_vc:
		case _p_slr_cv:
		case _p_plus_y_vv:
		case _p_plus_y_vc:
		case _p_minus_y_vv:
		case _p_minus_y_cv:
		case _p_times_y_vv:
		case _p_times_y_vc:
		case _p_div_y_vv:
		case _p_div_y_cv:
		case _p_div_y_vc:
		case _p_and_y_vv:
		case _p_and_y_vc:
		case _p_or_y_vv:
		case _p_or_y_vc:
		case _p_sll_y_vv:
		case _p_sll_y_vc:
		case _p_sll_y_cv:
		case _p_slr_y_vv:
		case _p_slr_y_vc:
		case _p_slr_y_cv:
		case _p_eq:
		case _p_arg:
		case _p_functor:
			break;
		default:
			/*
			 * This means there was more space allocated than
			 * what was really necessary 
			 */
			return;
		}
		/*
		 * Continue until either you reach the end of the code, or
		 * you find a procceed (we like to check for procceed or
		 * execute because for simple clauses you can have more code
		 * reserved than what you need) 
		 */
		/*
		 * } while ( AsCodePtr(code) < EndOfC-OpCodeSize+1 && op !=
		 * _procceed && op != _execute); 
		 */
	} while (TRUE);
#ifdef DEBUG_RESTORE2
	YP_fprintf(errout, "\n");
#endif
}

/*
 * Restores a group of clauses for the same predicate, starting with First
 * and ending with Last, First may be equal to Last 
 */
static void 
CleanClauses(CODEADDR First, CODEADDR Last)
{
	if (First == Last)
		RestoreClause(First);
	else {
		do {
			RestoreClause(First);
			First = NextClause(First);
		} while (First != Last);
		RestoreClause(Last);
	}
}

/*
 * This is used to make an hash table correct, after displacing its elements,
 * HCEnd should point to an area of free space, usually in the heap. The
 * routine is very dependent on the hash function used, and it destroys the
 * previous "hit" order 
 */
static void 
rehash(CELL *oldcode, int NOfE, int KindOfEntries)
{
	register CELL  *savep = HCEnd;
	CELL           *oldp = oldcode, *a;
	int             TableSize = NOfE - 1;
	register int    i;
	int             hash;
	CELL            WorkTerm, failplace = 0;
	int             rot;
	if (KindOfEntries == Atomics)
		rot = 4;
	else
		rot = 12;
	for (i = 0; i < NOfE; ++i) {
		if (*oldp == 0)
			failplace = *(oldp + 1);
		*savep++ = *oldp++;
		*savep++ = *oldp++;
	}
	oldp = oldcode;
	for (i = 0; i < NOfE; ++i) {
		*oldp++ = 0;
		*oldp++ = failplace;
	}
	savep = HCEnd;
	for (i = 0; i < NOfE; ++i) {
		if (*savep) {
			register Int    d;
			WorkTerm = *savep++;
			hash = (WorkTerm >> rot) & TableSize;
			a = oldcode + hash * 2;
			d = TableSize & (WorkTerm | 1);
			while (*a) {
				hash = (hash + d) & TableSize;
				a = oldcode + hash * 2;
			}
			*a++ = WorkTerm;
			*a = *savep++;
		} else
			savep += 2;
	}
}


/* Restores the indexing code, op by op */
/* This must be done only after all the clauses are restored */
static void 
RestoreIndex(CODEADDR Start)
{
#ifdef DEBUG_RESTORE2
	int             Size;
	CODEADDR        Last;
#endif
	register int    op;
	ptrs            code, s_c;
	AsCodePtr(s_c) = AsCodePtr(code) = Start;
	op = oldop(*AsOpCodePtr(code));
#ifdef DEBUG_RESTORE2
	Size = SizeOfBlock(Start);
	Last = Start + Size;
#endif
	do {
		/* if ( Last-AsCodePtr(code) < OpCodeSize) return; */
		op = oldop(*AsOpCodePtr(code));
#ifdef DEBUG_RESTORE2
		YP_fprintf(errout, "%d ", op);
#endif
		*AsOpCodePtr(code)++ = opcode((int) op);
		switch (op) {
		case _Ystop:
		case _Nstop:
#ifdef DEBUG_RESTORE2
			YP_fprintf(errout, "left OK\n");
#endif
			return;
		case _retry:
		case _trust_in:
		case _trust:
		case _retry_first:
		case _trust_first_in:
		case _trust_first:
		case _retry_tail:
		case _trust_tail_in:
		case _trust_tail:
		case _retry_head:
		case _trust_head_in:
		case _trust_head:
		case _jump_if_var:
			/* if (Last - AsCodePtr(code) < CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _try_me:
		case _retry_me:
		case _try_me0:
		case _retry_me0:
		case _try_me1:
		case _retry_me1:
		case _try_me2:
		case _retry_me2:
		case _try_me3:
		case _retry_me3:
		case _try_me4:
		case _retry_me4:
		case _try:
			/* if (Last - AsCodePtr(code) < CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			AsOpRegPtr(code)++;
			break;
		case _trust_me:
		case _trust_me0:
		case _trust_me1:
		case _trust_me2:
		case _trust_me3:
		case _trust_me4:
			/* if (Last - AsCodePtr(code) < CellSize) return; */
			if (*AsPtrCodePtr(code))
				*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _switch_on_nonv:
		case _switch_nv_list:
		case _switch_last:
		case _switch_l_list:
			/* if (Last - AsCodePtr(code) < 3*CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _switch_on_type:
		case _switch_on_head:
		case _switch_list_nl:
			/* if (Last - AsCodePtr(code) < 4*CellSize) return; */
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _go_on_cons:
			/* if (Last - AsCodePtr(code) < 3*CellSize) return; */
			if (!IsNumTerm(*AsTermPtr(code)))
				*AsAtomPtr(code) = AtomAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _go_on_func:
			/* if (Last - AsCodePtr(code) < 3*CellSize) return; */
			*AsFunctorPtr(code) = FuncAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		case _if_cons:{
				int             i = *AsOpRegPtr(code)++, j;
				/*
				 * if (Last - AsCodePtr(code) <
				 * CellSize+i*2*CellSize) return; 
				 */
				*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
				AsPtrCodePtr(code)++;
				for (j = 0; j < i; ++j) {
					if (!IsNumTerm(*AsTermPtr(code)))
						*AsAtomPtr(code) = AtomAdjust(*AsCellPtr(code));
					AsCellPtr(code)++;
					*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
					AsPtrCodePtr(code)++;
				}
			}
			break;
		case _if_func:{
				int             i = *AsOpRegPtr(code)++, j;
				/*
				 * if (Last - AsCodePtr(code) <
				 * CellSize+i*2*CellSize) return; 
				 */
				*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
				AsPtrCodePtr(code)++;
				for (j = 0; j < i; ++j) {
					*AsFunctorPtr(code) = FuncAdjust(*AsCellPtr(code));
					AsCellPtr(code)++;
					*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
					AsPtrCodePtr(code)++;
				}
			}
			break;
		case _switch_on_cons:{
				int             i = *AsOpRegPtr(code)++, j;
				CELL           *oldcode;
				/*
				 * if (Last - AsCodePtr(code) < i*2*CellSize)
				 * return; 
				 */
				oldcode = AsCellPtr(code);
				for (j = 0; j < i; ++j) {
					if ((*AsCellPtr(code) != 0)
					    && !IsNumTerm(*AsCellPtr(code)))
						*AsAtomPtr(code) = AtomAdjust(*AsCellPtr(code));
					AsCellPtr(code)++;
					*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
					AsPtrCodePtr(code)++;
				}
				rehash(oldcode, i, Atomics);
			}
			break;
		case _switch_on_func:{
				int             i, j;
				CELL            oldfunc, *oldcode;
				i = *AsOpRegPtr(code)++;
				/*
				 * if (Last - AsCodePtr(code) < i*2*CellSize)
				 * return; 
				 */
				oldcode = AsCellPtr(code);
				for (j = 0; j < i; ++j) {
					oldfunc = *AsCellPtr(code);
					if (oldfunc == 0)
						AsCellPtr(code)++;
					else {
						*AsFunctorPtr(code) = FuncAdjust(oldfunc);
						AsCellPtr(code)++;
					}
					*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
					AsPtrCodePtr(code)++;
				}
				rehash(oldcode, i, Funcs);
			}
			break;
		case _if_not_then:
			/* if (Last - AsCodePtr(code) < 3*CellSize) return; */
			if (!IsNumTerm(*AsTermPtr(code)))
				*AsAtomPtr(code) = AtomAdjust(*AsCellPtr(code));
			AsCellPtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			*AsPtrCodePtr(code) = CodeAdjust(*AsPtrCodePtr(code));
			AsPtrCodePtr(code)++;
			break;
		default:
			/* This means we had more space than what we needed */
			return;
		}
		/* } while (AsCodePtr(code) < Last ); */
	} while (TRUE);
#ifdef DEBUG_RESTORE2
	YP_fprintf(errout, "\n");
#endif
}

/*
 * Clean all the code for a particular predicate, this can get a bit tricky,
 * because of the indexing code 
 */
static void 
CleanCode(PredEntry *pp)
{
	CELL            flag;
	CODEADDR        FirstC, LastC;
	/* Init takes care of the first 2 cases */
	if (pp->ArityOfPE)
		pp->FunctorOfPred = FuncAdjust(pp->FunctorOfPred);
	else
		pp->FunctorOfPred = (Functor) AtomAdjust(pp->FunctorOfPred);
	if (pp->Owner.OwnerFile)
		pp->Owner.OwnerFile = AtomAdjust(pp->Owner.OwnerFile);
	if (pp->PredFlags & (CPredFlag | BasicPredFlag))
		return;
	if (pp->FirstClause)
		pp->FirstClause = CodeAdjust(pp->FirstClause);
	if (pp->LastClause)
		pp->LastClause = CodeAdjust(pp->LastClause);
	pp->CodeOfPred = CodeAdjust(pp->CodeOfPred);
	pp->TrueCodeOfPred = CodeAdjust(pp->TrueCodeOfPred);
	flag = pp->PredFlags;
	FirstC = pp->FirstClause;
	LastC = pp->LastClause;
	/* We just have a fail here */
	if (FirstC == NIL && LastC == NIL) {
	  pp->OpcodeOfPred = opcode(_undef_p);
	  pp->TrueCodeOfPred = pp->CodeOfPred =
	    (CODEADDR)(&(pp->OpcodeOfPred)); 
	  return;
	}
	if (flag & DynamicPredFlag) {
#ifdef	DEBUG_RESTORE2
		YP_fprintf(errout, "Correcting dynamic code\n");
#endif
		CleanClauses(FirstC, LastC);
		RestoreClause(pp->TrueCodeOfPred);
		return;
	}
	if ((flag & IndexedPredFlag)) {
#ifdef	DEBUG_RESTORE2
		YP_fprintf(errout, "Correcting indexed code from %lx to %lx\n", FirstC, LastC);
#endif
		CleanClauses(FirstC, LastC);
		RestoreIndex(pp->TrueCodeOfPred);
	} else
#ifdef	DEBUG_RESTORE2
	{
		YP_fprintf(errout, "at %lx Correcting clauses from %lx to %lx\n", *(OPCODE *) FirstC, FirstC, LastC);
#endif
		CleanClauses(FirstC, LastC);
#ifdef	DEBUG_RESTORE2
	}
#endif
}

/*
 * Restores all of the entries, for a particular atom, we only have problems
 * if we find code or data bases 
 */
static void 
RestoreEntries(PropEntry *pp)
{
	while (!EndOfPAEntr(pp)) {
		if (!(pp->KindOfPE & 0x8000))
			CleanCode((PredEntry *) pp);
		else if (pp->KindOfPE == ValProperty) {
			Term            tv = ((ValEntry *) pp)->ValueOfVE;
			if (IsAtomTerm(tv))
				((ValEntry *) pp)->ValueOfVE = (Term) AtomAdjust(tv);
		} else if (pp->KindOfPE == OpProperty);	/* Op */
#ifdef SFUNC
		else if (pp->KindOfPE == SFProperty) {
			CELL            tv = ((SFEntry *) pp)->NilValue;
			if (IsAtomTerm(tv))
				((SFEntry *) pp)->NilValue = AtomAdjust(tv);
		}
#endif
		else if ((pp->KindOfPE & 0xff00) == 0xbb00) {
			((FunctorEntry *) pp)->NameOfFE =
				AtomAdjust(((FunctorEntry *) pp)->NameOfFE);
		} else if ((pp->KindOfPE & 0xff00) &&
			   (pp->KindOfPE & 0x0020));	/* Expression */
		else if (pp->KindOfPE & 0x8000)
#ifdef DEBUG_RESTORE2
		{
			YP_fprintf(errout, "Correcting data base clauses\n");
#endif
			RestoreDB((DBEntry *) pp);
#ifdef DEBUG_RESTORE2
		}
#endif
		else
		  Error(SYSTEM_ERROR,TermNil,"restoring atoms");
		pp = RepProp(pp->NextOfPE);
	}
}

/* restore the atom entries which are invisible for the user */
static void 
RestoreInvisibleAtoms(void)
{
	AtomEntry      *at;

	at = RepAtom(*INVISIBLECHAIN);
	if (EndOfPAEntr(at))
		return;
#ifdef DEBUG_RESTORE2		/* useful during debug */
	do {
		YP_fprintf(errout, "Restoring %s\n", at->StrOfAE);
		RestoreEntries(RepProp(at->PropOfAE));
	}
#else
	do
		RestoreEntries(RepProp(at->PropOfAE));
#endif
	while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
}

/* restores the list of free space, with its curious structure */
static void 
RestoreFreeSpace(void)
{
	register BlockHeader *bpt, *bsz;
	bpt = BlockAdjust(OldFreeBlocks);
	FreeBlocks = bpt;
	while (bpt != NIL) {
		bsz = bpt->b_next = BlockAdjust(bpt->b_next);
		while (bsz != NIL) {
			bsz->b_next_size = BlockAdjust(bsz->b_next_size);
			bsz = bsz->b_next = BlockAdjust(bsz->b_next);
		}
		bpt = bpt->b_next_size = BlockAdjust(bpt->b_next_size);
	}
}

/*
 * This is the really tough part, to restore the whole of the heap 
 */
static void 
restore_heap(void)
{
	Atom           *HashPtr = HashChain;
	register int    i;
	for (i = 0; i < MaxHash; ++i) {
		if (*(HashPtr)) {
			AtomEntry      *at;
			at = RepAtom(*HashPtr);
#ifdef DEBUG_RESTORE2			/* useful during debug */
			do {
				YP_fprintf(errout, "Restoring %s\n", at->StrOfAE);
				RestoreEntries(RepProp(at->PropOfAE));
			}
#else
			do
				RestoreEntries(RepProp(at->PropOfAE));
#endif
			while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
		}
		HashPtr++;
	}
	/*
	 * for(i=0; i < NoOfModules; ++i) ModuleName[i] = (Term)
	 * AtomAdjust(ModuleName[i]); 
	 */
	RestoreInvisibleAtoms();
}


static Term 
AdjustAppl(register Term t)
{
	t = (Term) RepAppl(t);
	if (IsOldGlobal(t))
		return(AbsAppl((CELL *)GlobalAdjust(t)));
	else if (IsOldHeap(t))
		return (AbsAppl((CELL *)HeapAdjust(t)));
#ifndef MULTI_ASSIGNMENT_VARIABLES
#if SIZEOF_INT_P==4 && !(defined(YAPOR) && defined(ACOW))
	/* if the new tagging scheme we may have small objects that */
	/* look like appls. Just ignore them for the moment */ 
	else
	  Error(SYSTEM_ERROR,TermNil,
		"garbage %x found in stacks by restore/1",
		AbsAppl((CELL *)(t)));
#endif
#endif
	return (t);
}

static Term 
AdjustPair(register Term t)
{
	t = (Term) RepPair(t);
	if (IsOldGlobal(t))
		return (AbsPair((CELL *)GlobalAdjust(t)));
	else if (IsOldHeap(t))
		return (AbsPair((CELL *)HeapAdjust(t)));
#ifndef MULTI_ASSIGNMENT_VARIABLES
#if SIZEOF_INT_P==4 && !(defined(YAPOR) && defined(ACOW))
	else
	  Error(SYSTEM_ERROR,TermNil,
		"garbage %x found in stacks by restore/1",
		AbsPair((CELL *)t));
#endif
#endif
	return (t);
}

/* Restores the stacks, or the state of the computation */
static void 
RestoreState(void)
{
  register CELL  *pt = LCL0;
  register CELL   reg;
  register tr_fr_ptr ptt;
  /*
   * Now, I should convert the stacks, starting with the global and the
   * local 
   */
  /*
   * Maybe it would be better to go down->up and recognize the choice
   * points by the cells with small numbers 
   */
  while (pt >= ASP) {
    reg = *--pt;
#ifdef DEBUG_RESTORE5
    YP_fprintf(errout, "Restoring %lx in %lx until %lx\n", reg, pt, ASP);
#endif
    /* offsets in local stack: beats me if I know who puts
       them there */
    if (Signed(reg) < 0 && -Signed(reg) < (Int)OldHeapBase) {
      /* if(Signed(reg)<-256) YP_fprintf(YP_stderr,"[ neg %lx]\n",reg); */
      continue;
    }
    if (reg == 0x0)
      continue;
    if (IsVarTerm(reg)) {
      if (IsOldLocal(reg))
	*pt = LocalAdjust(reg);
      else if (IsOldCode(reg))
	*pt = HeapAdjust(reg);
      else if (IsOldGlobal(reg))
	*pt = GlobalAdjust(reg);
      else if (IsOldTrail(reg))
	*pt = TrailAdjust(reg);
    } else if (IsAtomTerm(reg))
      *pt = (CELL) AtomAdjust(reg);
    else if (IsApplTerm(reg))
      *pt = (CELL) AdjustAppl((Term) reg);
    else if (IsPairTerm(reg))
      *pt = (CELL) AdjustPair((Term) reg);
    else if (IsRefTerm(reg))
      *pt = (CELL) AtomAdjust(reg);
  }
  ptt = TR;
  /* To clean the trail is a simple business */
  while (ptt != (tr_fr_ptr)TrailBase) {
    reg = TrailTerm(ptt-1);
    ptt--;
    if (IsOldLocal(reg))
      TrailTerm(ptt) = LocalAdjust(reg);
    else if (IsOldGlobal(reg))
      TrailTerm(ptt) = GlobalAdjust(reg);
#ifdef MULTI_ASSIGNMENT_VARIABLES
    /* check it whether we are protecting a
       multi-assignment */
    else if (IsNonVarTerm(reg) < 0) {
      reg = (CELL)RepPair(reg);
      if (IsOldGlobal(reg) || IsOldLocal(reg)) {
	/* multi-assignment variable */
	/* we don't actually need to do anything here,
	   unless we are pointing to a mutable variable,
	   which itself is pointing to the trail. In this
	   case, we need to fix the mutable variable */
	if (IsOldGlobal(reg))
	  TrailTerm(ptt) = AbsPair((CELL *)GlobalAdjust(reg));
	else
	  TrailTerm(ptt) = AbsPair((CELL *)LocalAdjust(reg));
	/* If we do not fix the heap, we should be careful
	   to fix trail pointers in mutable variables */
	if (GDiff == 0 && IsOldTrail(*(CELL *)reg)) {
	  *(CELL *)reg = TrailAdjust(*(CELL *)reg);
	}
	/* now, go to the value field */
	reg = TrailTerm(ptt-1);
	ptt--;
	if (IsVarTerm(reg)) {
	  if (IsOldLocal(reg))
	    reg = LocalAdjust(reg);
	  else if (IsOldGlobal(reg))
	    reg = GlobalAdjust(reg);
	  else if (IsOldTrail(reg))
	    reg = TrailAdjust(reg);
	} else if (IsApplTerm(reg))
	  reg = AdjustAppl(reg);
	else if (IsPairTerm(reg))
	  reg = AdjustPair(reg);
	TrailTerm(ptt) = (Term) reg;
      }
    }
#endif
  }
  /*
   * to clean the heap now that functors are just variables pointing to
   * the code 
   */
  pt = (CELL *) H0;
  if (H0 != OldH0)
    while (pt < CellPtr(H)) {
      reg = *pt;
      if (reg == (CELL)FunctorDouble) {
	pt += 1+sizeof(double)/sizeof(CELL);
      } else if (reg == (CELL)FunctorLongInt) {
	pt += 2;
      } else if (IsVarTerm(reg)) {
	if (IsOldGlobal(reg))
	  *pt = GlobalAdjust(reg);
	else if (IsOldCode(reg))
	  *pt = HeapAdjust(reg);
#ifdef MULTI_ASSIGNMENT_VARIABLES
	else if (IsOldTrail(reg))
	  *pt = TrailAdjust(reg);
#endif
#if SBA
	else if (reg == 0)
	  {}
#endif
	else
	  Error(SYSTEM_ERROR, TermNil,
		"garbage %x found in global at %p by restore/1", reg, pt);
      } else if (IsAtomTerm(reg)) {
	*pt = (CELL) AtomAdjust(reg);
      } else if (IsApplTerm(reg)) {
	*pt = (CELL) AdjustAppl((Term) reg);
      } else if (IsPairTerm(reg)) {
	*pt = (CELL) AdjustPair((Term) reg);
      } else if (IsRefTerm(reg)) {
	*pt = (CELL) AtomAdjust(reg);
      }
#ifdef DEBUG_RESTORE5
      YP_fprintf(errout, "Restoring %lx in %lx to %lx\n", reg, pt, *pt);
#endif
      pt++;
    }
}

#ifdef DEBUG_RESTORE3
static void 
ShowEntries(pp)
	PropEntry      *pp;
{
  while (!EndOfPAEntr(pp)) {
    YP_fprintf(YP_stderr,"Estou a ver a prop %x em %x\n", pp->KindOfPE, pp);
    pp = RepProp(pp->NextOfPE);
  }
}

static void 
ShowAtoms()
{
  Atom           *HashPtr = HashChain;
  register int    i;
  for (i = 0; i < MaxHash; ++i) {
    if (*(HashPtr)) {
      AtomEntry      *at;
      at = RepAtom(*HashPtr);
      do {
	YP_fprintf(YP_stderr,"Passei ao %s em %x\n", at->StrOfAE, at);
	ShowEntries(RepProp(at->PropOfAE));
      } while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
    }
    HashPtr++;
  }
}

#endif /* DEBUG_RESTORE3 */

#include <stdio.h>

static int 
OpenRestore(char *s, int flag)
{

  /*  if (strcmp(s, StartUpFile) == 0)
      YP_fprintf(YP_stderr, "[ YAP version %s ]\n\n", version_number);*/
  CloseStreams(TRUE);
  if ((splfild = open_file(s, O_RDONLY)) < 0)
    return(FALSE);
  PrologMode = BootMode;
  if (check_header(flag) == FALSE)
    return(FALSE);
  YP_fprintf(YP_stderr, "[ Restoring file %s ]\n", s);
  LookupMode = FullLookup;
#ifdef DEBUG_RESTORE4
  /*
   * This should be another file, like the log file 
   */
  errout = YP_stderr;
#endif
  return(TRUE);
}

static void 
CloseRestore(void)
{
#ifdef DEBUG_RESTORE3
	ShowAtoms();
#endif
	close_file();
	LookupMode = NormalLookup;
	PrologMode = UserMode;
}

static void 
RestoreHeap(void)
{
  HCCode = HCStart = CellPtr(H);
  RestoreFreeSpace();
  PrepareAtoms();
  InitStaff();
  if (!PrepareCCode() || OldHeapBase != HeapBase) {
    restore_heap();
    InitBackCPreds();
#ifdef DEBUG_RESTORE1
    YP_fprintf(errout, "phase 1 done\n");
#endif
  }
}

/*
 * This function is called to know about the parameters of the last saved
 * state 
 */
int 
SavedInfo(char *FileName, int *ATrail, int *AStack, int *AHeap)
{
  char            pp[80];
  char		  msg[256];
  int             result;

  if ((splfild = open_file(FileName, O_RDONLY)) < 0)
    return(FALSE);
  /* skip the first line */
  do {
    myread(splfild, pp, 1);
  } while (pp[0] != 1);
  sprintf(msg, "YAPV%s", version_number);
  myread(splfild, pp, Unsigned(strlen(msg) + 1));
  if (strcmp(pp, msg) != 0) {
    if (PrologMode != BootMode)
      Error(SYSTEM_ERROR, TermNil,
	    "file %s not using a saved Prolog state", FileName);
    return(0);
  }
  result = get_cell();
  if (*AHeap)
    get_cell();
  else
    *AHeap = get_cell() / 1024;
  if (*AStack)
    get_cell();
  else
    *AStack = get_cell() / 1024;
  if (*ATrail)
    get_cell();
  else
    *ATrail = get_cell() / 1024;
  close_file();
  return (result);
}

/*
 * This function is called when wanting only to restore the heap and
 * associated registers 
 */
void 
RestoreCode(char *s)
{
  /* 15/5/98: This is not working!! */
  if ((splfild = open_file(s, O_RDONLY)) < 0)
    return;
  if (OpenRestore(s, DO_ONLY_CODE) == FALSE)
    return;
  get_heap_info();
  get_heap_regs();
  get_insts();
  get_hash();
  CopyCode();
  restore_heap_regs();
  RestoreHeap();
  CloseRestore();
}

static Int 
p_restore(void)
{
  char            s[FILENAME_MAX];

  Term t1 = Deref(ARG1);
  if (!GetName(s, t1)) {
    Error(TYPE_ERROR_LIST,t1,"restore/1");
    return(FALSE);
  }
  /* We are now restoring a prolog state */
  /* #ifdef YAPOR
  if (OpenRestore(s,DO_ONLY_CODE) == FALSE)
    return(FALSE);
  get_coded(DO_ONLY_CODE);
  restore_regs(DO_ONLY_CODE);
  #else*/
  if (OpenRestore(s,DO_EVERYTHING) == FALSE)
    return(FALSE);
  get_coded(DO_EVERYTHING);
  restore_regs(DO_EVERYTHING);
  /*#endif*/
  RestoreHeap();
  if (OldHeapBase != HeapBase || OldLCL0 != LCL0 || OldH0 != H0
      || OldTrailBase != TrailBase) {
    RestoreState();
#ifdef DEBUG_RESTORE2
    YP_fprintf(errout, "phase 2 done\n");
#endif
  }
  CloseRestore();
  InitUserExtensions();
  CurrentModule = 1;
/* commented by ricroc
#ifdef YAPOR
  GLOBAL_root_cp->cp_or_fr = GLOBAL_root_or_fr;
#endif
*/
  return (TRUE);
}

void 
InitSavePreds(void)
{
  Aenhanced = LookupAtom("$enhanced");
  InitCPred("$save", 1, p_save, SafePredFlag|SyncPredFlag);
  InitCPred("$restore", 1, p_restore, SyncPredFlag);
}
