/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		utilpreds.c						 *
* Last rev:	4/03/88							 *
* mods:									 *
* comments:	new utility predicates for YAP				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "@(#)utilpreds.c	1.3";
#endif

#include "Yap.h"
#include "Yatom.h"
#include "io.h"
#include "eval.h"

typedef struct {
	Term            old_var;
	Term            new_var;
}              *vcell;


STATIC_PROTO(Term  create_new_term, (Term));
STATIC_PROTO(CELL  vars_in_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_non_singletons_in_term, (void));
STATIC_PROTO(CELL  non_singletons_in_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_variables_in_term, (void));
STATIC_PROTO(Int   ground_complex_term, (CELL *, CELL *));
STATIC_PROTO(Int   p_ground, (void));
STATIC_PROTO(Int   p_cp_term, (void));
STATIC_PROTO(Int   var_in_complex_term, (CELL *, CELL *, Term));

#ifdef DEBUG
STATIC_PROTO(Int  p_force_trail_expansion, (void));
#endif /* DEBUG */

static vcell    tbase, ttop;

Term	FoundVarTerm, ReFoundVarTerm; /* check for trouble with restore */

static Term 
create_new_term(Term t)
{

  if (IsVarTerm(t)) {
    register vcell  vp = tbase;

    while (vp != ttop) {
      if ((vp)->old_var == t)
	return (vp->new_var);
      vp++;
    }
    vp = ttop++;
    vp->old_var = t;
    return (vp->new_var = MkVarTerm());
  } else if (IsPrimitiveTerm(t))
    return (t);
  else if (IsPairTerm(t)) {
    register Term   nh = create_new_term(HeadOfTerm(t));

    return (MkPairTerm(nh, create_new_term(TailOfTerm(t))));
  } else {
    Term            st[MaxArity];
    register Functor fun = FunctorOfTerm(t);
    register unsigned int    arity = ArityOfFunctor(fun), i;

    for (i = 0; i < arity; i++)
      st[i] = create_new_term(ArgOfTerm(i + 1, t));
    return (MkApplTerm(fun, arity, st));
  }
}

static Int 
p_cp_term(void)
{				/* copy_term(?,?) */
	Term            tf;
	tbase = ttop = (vcell) HeapTop + sizeof(CELL);
	tf = create_new_term(Deref(ARG1));
	return (unify(tf, ARG2));
}


#ifdef RATIONAL_TREES

#define UNWIND_CUNIF()                                        \
         while (visited < AuxSp) {                            \
            pt0 = (CELL *)visited[0];                         \
            *pt0 = visited[1];                                \
            visited += 2;                                     \
         }

#else
#define UNWIND_CUNIF()
#endif

static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
  register tr_fr_ptr TR0 = TR;
  CELL *InitialH = H;
  CELL output = AbsPair(H);
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
	
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (f == FunctorDelay
#ifdef YAP_ARRAYS
	      || f == FunctorArrayAccess
#endif
	      ) {
	    d0 = Deref(d0);
	    goto continue_loop;
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    /* do or pt2 are unbound  */
    *ptd0 = TermNil;
    /* leave an empty slot to fill in later */
    H[1] = AbsPair(H+2);
    H += 2;
    H[-2] = (CELL)ptd0;
    /* next make sure noone will see this as a variable again */ 
    TrailTerm(TR++) = (CELL)ptd0;
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }

  UNWIND_CUNIF();
  /* we've done it */
  /* restore our nice, friendly, term to its original state */
  if (TR != TR0) {
    do {
      CELL *p = (CELL *)TrailTerm(--TR);
      RESET_VARIABLE(p);
    } while (TR != TR0);
  }
  if (H != InitialH) {
    /* close the list */
    RESET_VARIABLE(H-1);
    unify((CELL)(H-1),ARG2);
    return(output);
  } else {
    return(ARG2);
  }
}
 
static Int 
p_variables_in_term()		/* variables in term t		 */
{
        Term t = Deref(ARG1);
	Term out;
	if (IsVarTerm(t)) {
	  out = AbsPair(H);
	  H += 2;
	  RESET_VARIABLE(H-2);
	  RESET_VARIABLE(H-1);
	  unify((CELL)(H-2),ARG1);
	  unify((CELL)(H-1),ARG2);
	}  else if (IsPrimitiveTerm(t)) 
	  out = ARG2;
	else if (IsPairTerm(t)) {
	  out = vars_in_complex_term(RepPair(t)-1,
				     RepPair(t)+1);
	}
	else out = vars_in_complex_term(RepAppl(t),
					RepAppl(t)+
					ArityOfFunctor(FunctorOfTerm(t)));
	return(unify(ARG3,out));
}


static Term non_singletons_in_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
  register tr_fr_ptr TR0 = TR;
  CELL *InitialH = H;
  CELL output = AbsPair(H);
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (f == FunctorDelay
#ifdef YAP_ARRAYS
	      || f == FunctorArrayAccess
#endif
	      ) {
	    d0 = Deref(d0);
	    goto continue_loop;
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      } else if (d0 == FoundVarTerm) {
	CELL *pt2 = pt0;
	while(IsVarTerm(*pt2))
	  pt2 = (CELL *)(*pt2);
	H[1] = AbsPair(H+2);
	H += 2;
	H[-2] = (CELL)pt2;
	*pt2 = ReFoundVarTerm;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    /* do or pt2 are unbound  */
    *ptd0 = FoundVarTerm;
    /* next make sure we can recover the variable again */ 
    TrailTerm(TR++) = (CELL)ptd0;
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }

  UNWIND_CUNIF();
  /* we've done it */
  /* restore our nice, friendly, term to its original state */
  if (TR != TR0) {
    do {
      CELL *p = (CELL *)TrailTerm(--TR);
      RESET_VARIABLE(p);
    } while (TR != TR0);
  }
  if (H != InitialH) {
    /* close the list */
    RESET_VARIABLE(H-1);
    unify((CELL)(H-1),ARG2);
    return(output);
  } else {
    return(ARG2);
  }
}
 
static Int 
p_non_singletons_in_term()		/* non_singletons in term t		 */
{
        Term t = Deref(ARG1);
	Term out;
	if (IsVarTerm(t)) {
	  out = MkPairTerm(t,ARG2);
	}  else if (IsPrimitiveTerm(t)) 
	  out = ARG2;
	else if (IsPairTerm(t)) {
	  out = non_singletons_in_complex_term(RepPair(t)-1,
				     RepPair(t)+1);
	}
	else out = non_singletons_in_complex_term(RepAppl(t),
					RepAppl(t)+
					ArityOfFunctor(FunctorOfTerm(t)));
	return(unify(ARG3,out));
}

static Int ground_complex_term(register CELL *pt0, register CELL *pt0_end)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	if (f == FunctorDelay
#ifdef YAP_ARRAYS
	      || f == FunctorArrayAccess
#endif
	    ) {
	  d0 = Deref(d0);
	  goto continue_loop;
	}
#endif
	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    UNWIND_CUNIF();
    return(FALSE);
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }
  UNWIND_CUNIF();
  return(TRUE);
}
 
static Int 
p_ground()			/* ground(+T)		 */
{
        Term t = Deref(ARG1);

	if (IsVarTerm(t)) {
	  return(FALSE);
	}  else if (IsPrimitiveTerm(t)) {
	  return(TRUE);
	} else if (IsPairTerm(t)) {
	  return(ground_complex_term(RepPair(t)-1,
				     RepPair(t)+1));
	} else {
	  Functor fun = FunctorOfTerm(t);
	  if (IsExtensionFunctor(fun))
	    return(TRUE);
	  else return(ground_complex_term(RepAppl(t),
					  RepAppl(t)+
					  ArityOfFunctor(fun)));
	}
}

static Int var_in_complex_term(register CELL *pt0, register CELL *pt0_end, Term t)
{

  register CELL **to_visit = (CELL **)(HeapTop + sizeof(CELL));
#ifdef RATIONAL_TREES
  register CELL *visited = AuxSp;
#endif

 loop:
  while (pt0 < pt0_end) {
    register CELL d0;
    register CELL *ptd0;
    ++ pt0;
    ptd0 = pt0;
    d0 = *ptd0;
#ifdef COROUTINING
  continue_loop:
#endif
    deref_head(d0, vars_in_term_unk);
  vars_in_term_nvar:
    {
      if (IsPairTerm(d0)) {
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	pt0 = RepPair(d0) - 1;
	pt0_end = RepPair(d0) + 1;
      } else if (IsApplTerm(d0)) {
	register Functor f;
	register CELL *ap2;
	/* store the terms to visit */
	ap2 = RepAppl(d0);
	f = (Functor)(*ap2);

	if (IsExtensionFunctor(f)) {
#ifdef COROUTINING
	  if (f == FunctorDelay
#ifdef YAP_ARRAYS
	      || f == FunctorArrayAccess
#endif
	      ) {
	    d0 = Deref(d0);
	    goto continue_loop;
	  }
#endif

	  continue;
	}
#ifdef RATIONAL_TREES
	/* now link the two structures so that no one else will */
	/* come here */
	visited -= 2;
	visited[0] = (CELL) pt0;
	visited[1] = *pt0;
	*pt0 = TermNil;
#endif
	/* store the terms to visit */
	if (pt0 < pt0_end) {
	  to_visit[0] = pt0;
	  to_visit[1] = pt0_end;
	  to_visit += 2;
	}
	d0 = ArityOfFunctor(f);
	pt0 = ap2;
	pt0_end = ap2 + d0;
      }
      continue;
    }
    

    deref_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
    if ((CELL)ptd0 == t) { /* we found it */
      UNWIND_CUNIF();
      return(TRUE);
    }
  }
  /* Do we still have compound terms to visit */
  if (to_visit > (CELL **)(HeapTop + sizeof(CELL))) {
    to_visit -= 2;
    pt0 = to_visit[0];
    pt0_end = to_visit[1];
    goto loop;
  }
  UNWIND_CUNIF();
  return(FALSE);
}
 
Int 
var_in_term(Term v, Term t)		/* variables in term t		 */
{

  if (IsPrimitiveTerm(t)) 
    return(FALSE);
  else if (IsPairTerm(t)) {
    return(var_in_complex_term(RepPair(t)-1,
			       RepPair(t)+1, v));
  }
  else return(var_in_complex_term(RepAppl(t),
				  RepAppl(t)+
				  ArityOfFunctor(FunctorOfTerm(t)),v));
}



#ifdef DEBUG
static Int
p_force_trail_expansion()
{
  Int i = IntOfTerm(Deref(ARG1))*1024, j = 0;
  tr_fr_ptr OTR = TR;

  for (j = 0; j < i; j++) {
    TrailTerm(TR) = 0;
    TR++;
  }
  TR = OTR;

  return(TRUE);
}
#endif /* DEBUG */

void InitUtilCPreds(void)
{
  Atom AtomFoundVar = LookupAtom("$I_FOUND_A_VARIABLE");
  Atom AtomReFoundVar = LookupAtom("$I_FOUND_THE_VARIABLE_AGAIN");

  FoundVarTerm = MkAtomTerm(AtomFoundVar); 
  ReFoundVarTerm = MkAtomTerm(AtomReFoundVar);
  /* make sure no one else can use these two atoms */
  ReleaseAtom(AtomFoundVar);
  ReleaseAtom(AtomReFoundVar);

  InitCPred("copy_term", 2, p_cp_term, SafePredFlag);
  InitCPred("ground", 1, p_ground, SafePredFlag);
  InitCPred("$variables_in_term", 3, p_variables_in_term, SafePredFlag);
  InitCPred("$non_singletons_in_term", 3, p_non_singletons_in_term, SafePredFlag);
#ifdef DEBUG
  InitCPred("$force_trail_expansion", 1, p_force_trail_expansion, SafePredFlag);
#endif
}

