/*************************************************************************
*									 *
*	 Yap Prolog 							 *
*									 *
*	Yap Prolog Was Developed At Nccup - Universidade Do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa And Universidade Do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		Yap.C							 *
* Last Rev:								 *
* Mods:									 *
* Comments:	Yap'S Main File						 *
*									 *
*************************************************************************/
/* static char SccsId[] = "X 4.1.19"; */

#include "Yap.h"
#include "Yatom.h"
#include "Foreign.h"
#include "io.h"
#if defined(YAPOR) || defined(TABLING)
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif	/* YAPOR || TABLING */

#include <stdarg.h>
#include <stdlib.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef PROTECTED
#include <sys/time.h>
#endif /* Protected */

STATIC_PROTO(int mygetc, (Int *));
STATIC_PROTO(void do_boot, (char *));
STATIC_PROTO(void do_goal,(Term));

#ifndef LIGHT
STD_PROTO (void exit, (int));
#endif

#ifdef LIGHT
#include <unix.h>
int
  _main (int, char **);
#else
STD_PROTO (int main, (int, char **));
#endif

#if SIZEOF_INT_P==4
#if defined(YAPOR) || defined(TABLING)
#define MinTrailSpace	        192
#define DefTrailSpace		192
#define DefStackSpace	        1200
#define DefHeapSpace	        1200
#else
#define MinTrailSpace	        128
#define DefTrailSpace		128
#define DefStackSpace		800
#define DefHeapSpace		800
#endif /* YAPOR || TABLING */
#else
#if defined(YAPOR) || defined(TABLING)
#define MinTrailSpace	        384
#define DefTrailSpace		384
#define DefStackSpace	        2400
#define DefHeapSpace	        2400
#else
#define MinTrailSpace	        256
#define DefTrailSpace		256
#define DefStackSpace		1600
#define DefHeapSpace		1600
#endif /* YAPOR || TABLING */
#endif

#ifdef M_WILLIAMS
long _stksize = 32000;
#endif

static int YapInited = 0;

extern char StartUpFile[], BootFile[];

/************ variables	concerned with Error Handling *************/
sigjmp_buf         RestartEnv;	/* used to restart after an abort execution */

#ifdef lint
/* VARARGS1 */
#endif

static int pos = 0, eof_found = 0;

static FILE *bootfile;

#ifdef DEBUG
int output_msg = FALSE;
#endif

static int
mygetc (charpos)
     Int *charpos;
{
  int ch;
  if (eof_found)
    return (-1);
  ch = YP_getc (bootfile);
  if (ch == -1)
    eof_found = TRUE;
  *charpos = ++pos;
#ifdef MPW
  if (ch == '\n')
    ch = 10;
#endif
  return (ch);
}

void
do_goal (Goal)
     Term Goal;
{
  Term *CodeAdr;
  Atom name;
  int Arity, i;
  Int rcode;
  Atom at;
  PredEntry *undefpe;
  LookupMode = FullLookup;
  at = LookupAtom ("$undefp");
  LookupMode = NormalLookup;
  undefpe = RepPredProp (GetPredProp (at, 1));
  UndefCode = (CELL *) & (undefpe->CodeOfPred);
  if (IsAtomTerm (Goal))
    name = AtomOfTerm (Goal), Arity = 0;
  else
    {
      name = NameOfFunctor (FunctorOfTerm (Goal));
      Arity = ArityOfFunctor (FunctorOfTerm (Goal));
    }
  CodeAdr = ((Term *) (RepPredProp (PredProp (name, Arity))->CodeOfPred));
  for(i=1; i<=Arity; ++i)
    XREGS[i] = ArgOfTerm (i, Goal);
  /* create an initial pseudo environment so that when garbage
     collection is going up in the environment chain it doesn't get
     confused */
  ENV = YENV = ASP;
  YENV[E_CP] = (CELL)YESCODE;
  YENV[E_CB] = (CELL)NULL;
  YENV[E_E] = (CELL)NULL;
  ASP -= EnvSizeInCells;
  /* and now create a pseudo choicepoint for much the same reasons */
  CP = (yamop *) YESCODE;
  B = (choiceptr)ASP;
  B--;
  B->cp_h = H;
  B->cp_tr = TR;
  B->cp_cp = CP;
  B->cp_ap = (yamop *) NOCODE;
  B->cp_env = YENV;
  B->cp_b = (choiceptr) NULL;
#ifdef DEPTH_LIMIT
  B->cp_depth = DEPTH;
#endif /* DEPTH_LIMIT */
#if COROUTINING
  WokenGoals = NewTimedVar(TermNil);
  WakeUpCode = NULL;
#ifdef MULTI_ASSIGNMENT_VARIABLES
  MutableList = NewTimedVar(TermNil);
#endif
#endif
/* commented by ricroc
#if defined(YAPOR) || defined(TABLING)
  optyap_on();
#endif
*/
  YENV = ASP = (CELL *)B;
  HB = H;
#if defined(YAPOR) && defined(SBA)
  BB = B;
#endif
  YENV[E_CB] = Unsigned (B);
  P = (yamop *) CodeAdr;
  S = CellPtr (&(RepPredProp (PredProp (name, Arity))->StateOfPred));	/* A1 mishaps */
#ifdef DEBUG
  if (output_msg)
    YP_fprintf(YP_stderr,"Entering absmi\n");
#endif
  if (TrailTop - HeapTop < 2048)
    {
      PrologMode = BootMode;
      Error(SYSTEM_ERROR,TermNil,
	    "unable to boot because of too little heap space");
    }
  /* PlPutc(0,'a'); PlPutc(0,'\n'); */
  if (sigsetjmp (RestartEnv, 1) != 0) {
    int depth;
    /* otherwise, SetDBForThrow will fail entering critical mode */
    PrologMode = UserMode;
    /* find out where to cut to */
    depth = SetDBForThrow(MkAtomTerm(LookupAtom("abort")));
    if (depth == 0) {
      Error(SYSTEM_ERROR, TermNil, "database entry for throw corrupted");
    }
    /* make the abstract machine jump where we want them to jump to */
    B = (choiceptr)(LCL0-depth);
    P = (yamop *)FAILCODE;
  }
again:
  rcode = absmi ( 0 );
  if (rcode == 2)
    {				/* Abort("[ Fatal error: out of stack space
				 * ]"); */
    }
  else if (rcode == 3)
    {				/* Abort("[ Fatal error: out of heap space
				 * ]"); */
      if (!growheap()) {
	exit(1);
      }
     goto again;
    }
  else if (rcode == 4)
    {				/* Abort("[ Fatal error: out of trail space
				 * ]"); */
      goto again;
    }
}

static void
do_boot (bootfilename)
     char *bootfilename;
{
  Term t;
  TokEntry *tokstart;
  TR = (tr_fr_ptr)TrailBase;
  H = H0 = CellPtr (GlobalBase);
  LCL0 = ASP = CellPtr (LocalBase);
#ifdef DEPTH_LIMIT
  DEPTH = MkIntTerm(1000000);
#endif
  /* consult boot.pl */
  bootfile = YP_fopen (bootfilename, "r");
  if (bootfile == NIL)
    {
#if HAVE_STRERROR
      Error(SYSTEM_ERROR, TermNil, "unable to boot using %s, %s", bootfilename, strerror(errno));
#else
      Error(SYSTEM_ERROR, TermNil, "unable to boot using %s, %s", strerror(errno));
#endif
      exit(1);
    }
  PutValue (AtomIndex, MkAtomTerm (AtomTrue));
  while (!eof_found)
    {
      pos = 0;
      ParserAuxSp = ASP - 64;
      tokptr = toktide = tokstart = tokenizer (FunAdr (mygetc));
      if (eof_found)
	break;
      if (ErrorMessage)
	{
	  YP_fprintf(YP_stderr, "%s\n", ErrorMessage);

	  break;
	}
      t = Parse ();
      if (t == 0 || IsVarTerm (t))
	{
	  YP_fprintf(YP_stderr, "** syntax error**\n");
	  syntax_error (tokstart);
	}
      else if (t == MkAtomTerm (AtomTrue))
	{
	  exit_yap (0, NIL);
	}
      else if (IsPairTerm (t))
	Error(FATAL_ERROR, TermNil, "consult not allowed in boot.pl");
      else if (IsApplTerm (t) && FunctorOfTerm (t) == FunctorQuery) 
	{ 
	  do_goal (ArgOfTerm (1, t));
        }
      else
	{
	  CODEADDR codeadr = cclause (t,0);
	  if (ErrorMessage)
	    YP_fprintf(YP_stderr, ErrorMessage);
	  else
	    addclause (t, codeadr, TRUE);
	  if (ErrorMessage)
	    YP_fprintf(YP_stderr, ErrorMessage);
	}
    }
  YP_fclose (bootfile);
#ifdef DEBUG
  if (output_msg)
    YP_fprintf(YP_stderr,"Boot loaded\n");
#endif
}

#ifdef PROTECTED
static
after_time ()
{
  struct tm *ttm;
  long nday, time_t;

  time_t = time ((long *) 0);
  ttm = localtime (&time_t);
  nday = ttm->tm_year * 1000 + ttm->tm_yday;

  if ((nday < (Ym * 1000 + Dm)) || (nday > (YM * 1000 + DM)))
    {
      YP_fprintf(YP_stderr, "[ Licence period expired ]\n");
      exit (1);
    }
  if ((nday < (Ym * 1000 + Dm)) || (nday > (YM * 1000 + DM - 30)))
    {
      YP_fprintf(YP_stderr,
	       "[ Warning: your licence period will expire in %d days: please renew it! ]\n",
	       YM * 1000 + DM - nday);
    }
}

#endif /* PROTECTED */

int in_linbo = 1;

#ifdef LIGHT
int
_main (argc, argv)
#else
int
main (argc, argv)
#endif
     int argc;
     char **argv;
{
  int i, Heap = 0, Stack = 0, Trail = 0, Boot = FALSE, *ssize;
  char *p, *filename = NULL;
  char ch;
#if defined(YAPOR) || defined(TABLING)
  int aux_number_workers;
  int aux_scheduler_loop;
  int aux_delayed_release_load;
  int fd_map;
  char MapFile[20];
#endif /* YAPOR || TABLING */

#if PUSH_REGS
    /* In this case we need to initialise the abstract registers */
    REGSTORE standard_regs;
    regp = &standard_regs;
    /* the emulator will eventually copy them to its own local
       register array, but for now they exist */
#endif /* PUSH_REGS */


#if defined(YAPOR) || defined(TABLING)
  /* initializations */
#ifdef YAPOR
  worker_id = 0;
#endif /* YAPOR */
  aux_number_workers = 1;
  aux_scheduler_loop = 10;
  aux_delayed_release_load = 3;
  GLOBAL = NULL;
  strcpy(MapFile,"/tmp/mapfile");
  itos(getpid(), &MapFile[12]);
#endif /* YAPOR || TABLING */

  version_number = YAP_VERSION;

  /*
	 * proccess command line arguments: valid switches are: -b    boot -s
	 * stack area size (K) -h    heap area size -a    aux stack size -e
	 * emacs_mode -m    reserved memory for alloc IF DEBUG -p    if you
	 * want to check out startup IF MAC -mpw  if we are using the mpw
	 * shell
	 */
  yap_args = argv;
  yap_argc = argc;
  while (--argc > 0)
    {
      p = *++argv;
      if (*p == '-')
	switch (*++p)
	  {
	  case 'b':
	    Boot = 1;
	    break;
          case '?':
            YP_fprintf(YP_stderr,"\n[ Valid switches for command line arguments: ]\n");
            YP_fprintf(YP_stderr,"  -?   Shows this screen\n");
            YP_fprintf(YP_stderr,"  -b   Boot file (%s)\n", StartUpFile);
            YP_fprintf(YP_stderr,"  -h   Heap area in Kbytes (%d)\n", DefHeapSpace);
	    YP_fprintf(YP_stderr,"  -s   Stack area in Kbytes (%d)\n", DefStackSpace);
	    YP_fprintf(YP_stderr,"  -t   Aux/trail area in Kbytes (%d)\n", DefTrailSpace);
#ifdef YAPOR
            YP_fprintf(YP_stderr,"  -w   Number of workers (%d)\n", aux_number_workers);
            YP_fprintf(YP_stderr,"  -l   Loop scheduler executions before look for hiden shared work (%d)\n", aux_scheduler_loop);
            YP_fprintf(YP_stderr,"  -d   Value of delayed release of load (%d)\n", aux_delayed_release_load);
#endif /* YAPOR */
            YP_fprintf(YP_stderr,"\n");
            exit(0);
#ifdef YAPOR
	  case 'w':
	    ssize = &aux_number_workers;
	    goto GetSize;
          case 'l':
            ssize = &aux_scheduler_loop;  
	    goto GetSize;
          case 'd':
            ssize = &aux_delayed_release_load;
	    goto GetSize;
#endif /* YAPOR */
#ifdef HAVE_SOCKET
          case 'c':          /* running as client */
	    {
	      char *host, *p1;
	      long port;
	      char *ptr;

	      host = *++argv;
	      argc--;
	      if (host != NULL && host[0] == '-')
		Error(FATAL_ERROR,TermNil,
		      "sockets must receive host to connect to");
	      p1 = *++argv;
	      argc--;
	      if (p1[0] == '-')
		Error(FATAL_ERROR,TermNil,
		      "sockets must receive port to connect to");
	      port = strtol(p1, &ptr, 10);
	      if (ptr == NULL || ptr[0] != '\0')
		Error(FATAL_ERROR,TermNil,
		      "port argument to socket must be a number");
	      init_socks(host,port);
	    }
	    break;
#endif
#ifdef EMACS
	  case 'e':
	    emacs_mode = TRUE;
	    {
	      YP_File fd;
	      strcpy (emacs_tmp, ++p);
	      if ((fd = YP_fopen (emacs_tmp, "w")) == NIL)
		YP_fprintf(YP_stderr, "[ Warning: unable to communicate with emacs: failed to open %s ]\n", emacs_tmp);
	      YP_fclose (fd);
	      unlink (emacs_tmp);
	      p = *++argv;
	      --argc;
	      strcpy (emacs_tmp2, p);
	      if ((fd = YP_fopen (emacs_tmp2, "w")) == NIL)
		YP_fprintf(YP_stderr, "Unable to communicate with emacs: failed to open %s\n", emacs_tmp2);
	      YP_fclose (fd);
	      unlink (emacs_tmp2);
	    }
	    break;
#endif /* EMACS */
#ifdef MPWSHELL
	  case 'm':
	    if (*++p == 'p' && *++p == 'w' && *++p == '\0')
	      mpwshell = TRUE;
	    break;
#endif
	  case 's':
	    ssize = &Stack;
	    goto GetSize;
	  case 'h':
	    ssize = &Heap;
	    goto GetSize;
	  case 't':
	    ssize = &Trail;
	  GetSize:
	    if (*++p == 0)
	      {
		if (argc > 0)
		  --argc, p = *++argv;
		else
		  {
		    Error(FATAL_ERROR, TermNil, "missing size in flag");
		    exit (1);
		  }
	      }
	    i = 0;
	    while ((ch = *p++) >= '0' && ch <= '9')
	      i = i * 10 + ch - '0';
	    if (ch)
	      {
		Error(FATAL_ERROR, TermNil, "illegal size specification");
		exit (1);
	      }
	    *ssize = i;
	    break;
#ifdef DEBUG
	  case 'p':
	    output_msg = TRUE;
	    break;
	  case 'q':
	    {
	      extern int tr_heap;
	      tr_heap = 340;
	    }
	    break;
#endif
	  case '-':
	    /* skip remaining arguments */
	    argc = 1;
	    break;
	  default:
	    Error(FATAL_ERROR, TermNil, "unknown switch -%c", *p);
	    exit (1);
	  }
      else {
	filename = p;
      }
    }

  /* init memory */
  if (Boot)
    {
      if (!Trail)
	Trail = DefTrailSpace;
      if (!Stack)
	Stack = DefStackSpace;
      if (!Heap)
	Heap = DefHeapSpace;
    }
  else
    {
      if (!filename)
	filename = StartUpFile;
      if (SavedInfo (filename, &Trail, &Stack, &Heap) != 1) {
#if HAVE_STRERROR
	if (errno)
	  Error(FATAL_ERROR, TermNil, "could not open saved state %s, %s", filename, strerror(errno));
	else
#endif
	  Error(FATAL_ERROR, TermNil, "%s is not a saved state", filename);
	exit (1);
      }
    }
#ifdef PROTECTED
  after_time ();
#endif
  if (Trail < MinTrailSpace)
    Trail = MinTrailSpace;
  InitSysbits ();
#ifdef SIMICS
  YP_fprintf(YP_stdout,"\n");
#endif /* SIMICS */
#if defined(YAPOR) || defined(TABLING)
  /* starting message */
#ifdef YAPOR
  if (aux_number_workers > MAX_WORKERS)
    abort_optyap("excessive number of workers");
#ifdef TABLING
  optyap_message_type0("Starting OPTYap with %d worker%s ... wait a moment",
	               aux_number_workers, aux_number_workers == 1 ? "":"s");
#else
  optyap_message_type0("Starting YapOr with %d worker%s ... wait a moment",
	               aux_number_workers, aux_number_workers == 1 ? "":"s");
#endif /* TABLING */
#else /* TABLING */
  optyap_message_type0("Starting YapTab ... wait a moment");
#endif /* YAPOR */
  /* memory mapping */ 
  if ((fd_map = open(MapFile, O_RDWR|O_CREAT|O_TRUNC, 0666)) <0)
    abort_optyap("open error in function main");
  map_memory(Heap, Stack, Trail, fd_map, aux_number_workers);
  /* global initializations */ 
  init_global(aux_number_workers, aux_scheduler_loop, aux_delayed_release_load);
  alloc_all_optyap_structs();
  init_signals();
#else /* Yap */
  InitMemory (Trail, Heap, Stack);
#endif /* YAPOR || TABLING */
  /* Init signal handling and time */
  Init ();
  InitUserExtensions ();
  YapInited = 1;
  /* check for boot */
  AuxSp = (CELL *)AuxTop;
  TR = (tr_fr_ptr)TrailBase;
  H = H0 = (CELL *) GlobalBase;
  LCL0 = ASP = (CELL *) LocalBase;
  if (Boot)
    {
      in_linbo = 0;
      do_boot (filename ? filename : BootFile);
      PutValue (AtomLive, MkAtomTerm (AtomTrue));
      PrologMode = UserMode;
    }
  else
    {
      /* assert clause:    '$live' :- '$restore'(...)   */
      Atom Restore = LookupAtom ("$restore");
      Term Goal = StringToList (filename);
      Term clause;
      Term t[2];
      CODEADDR cc;
      t[0] = MkAtomTerm (AtomLive);
      t[1] = MkApplTerm (MkFunctor (Restore, 1), 1, &Goal);
      cc = cclause (clause = MkApplTerm (FunctorAssert, 2, t), 0);
      if (ErrorMessage) {
	Error(Error_TYPE, Error_Term, ErrorMessage);
      } else
	addclause (clause, cc, 0);
    }
  /* now start up prolog by executing $live */
  TR = (tr_fr_ptr)TrailBase;
  /* we now enter normal execution mode */
  STATIC_PREDICATES_MARKED = FALSE;
    
#if PUSH_REGS
  /* Guarantee that after a longjmp we go back to the original abstract
     machine registers */
  regp = &standard_regs;
#endif /* PUSH_REGS */
  CurrentModule = 1;
  PutValue (AtomBreak, MkIntTerm (0));
  AuxSp = (CELL *)AuxTop;
  TR = (tr_fr_ptr)TrailBase;
  H = H0 = (CELL *) GlobalBase;
  LCL0 = ASP = (CELL *) LocalBase;
#ifdef DEPTH_LIMIT
  DEPTH = MkIntTerm(1000000);
#endif
  in_linbo = 0;

#if defined(YAPOR) || defined(TABLING)
  make_root_frames();
#ifdef YAPOR
  init_workers(fd_map);
#endif /* YAPOR */
  init_local();
  close(fd_map);
#ifdef YAPOR
  if (worker_id != 0) {
volatile int vsc_vol = 0;

    printf("I have been here\n"); 
    while(vsc_vol==0);

    P = GETWORK_FIRST_TIME;
    absmi(0);
    abort_optyap("unexpected exit in function main");
  }
#endif /* YAPOR */
#endif /* YAPOR || TABLING */

  while (GetValue (AtomLive) != MkAtomTerm (AtomFalse))
    {
      do_goal (MkAtomTerm (AtomLive));
      if (PrologMode & BootMode)
	{
	  exit_yap (1, NIL);
	}
    }
  exit_yap (0, NIL);
  return (0);
}

void
exit_yap (value, msg)
     int value;
     char *msg;
{
  if (msg != NIL) {
    if (value == 0) 
      /* no error */
      YP_fprintf(YP_stderr, msg);
    else
      /* error */
      YP_fprintf(YP_stderr, "[ YAP unrecoverable error: %s ]\n", msg);
  }
  ShutdownYAP (value);
}

